from typing import List, Optional, Tuple

from boa3.model.builtin.builtin import Builtin
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class Concat(BinaryOperation):
    """
    A class used to represent a string concatenation operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.str, Type.bytes, Builtin.ByteString]

    def __init__(self, left: IType = Type.str, right: IType = None):
        self.operator: Operator = Operator.Plus
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        if left == right and any(valid_type.is_type_of(left) for valid_type in self._valid_types):
            return True

        left_valid_type = self._get_valid_type(left)
        if left_valid_type is None:
            return False

        right_valid_type = self._get_valid_type(right)

        return (right_valid_type is not None
                and (left_valid_type == right_valid_type
                     or left_valid_type.is_type_of(right_valid_type)
                     or right_valid_type.is_type_of(left_valid_type))
                )

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            left_internal = self._get_valid_type(left)
            right_internal = self._get_valid_type(right)

            if left_internal.is_type_of(Builtin.ByteString):
                # for bytestring + str the result is str and bytestring + bytes is bytes
                return right
            if right_internal.is_type_of(Builtin.ByteString):
                return left

            if right.is_type_of(left):
                return right
            if left.is_type_of(right):
                return left

            return left_internal
        else:
            return Type.none

    def _get_valid_type(self, operator_type: IType) -> Optional[IType]:
        return next((valid_type for valid_type in self._valid_types
                     if valid_type.is_type_of(operator_type)), None)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.CAT, b''),
                (Opcode.CONVERT, Type.str.stack_item)
                ]



from __future__ import annotations

import base64
import enum
from typing import Any, Dict

from boa3.neo.vm.type.String import String
from boa3_test.tests.test_classes.transactionattribute import TransactionAttribute, TransactionAttributeType


class OracleResponse(TransactionAttribute):
    def __init__(self, request_id: int, code: OracleResponseCode, result: bytes):
        super().__init__(TransactionAttributeType.OracleResponse)
        self._id: int = request_id
        self._response_code: OracleResponseCode = code
        self._result: bytes = result

    def to_json(self) -> Dict[str, Any]:
        json = super().to_json()
        json.update({
            'id': self._id,
            'code': self._response_code.value,
            'result': String.from_bytes(base64.b64encode(self._result))
        })
        return json

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> OracleResponse:
        return cls(request_id=json['id'],
                   code=json['code'],
                   result=base64.b64decode(json['result']))

    def __eq__(self, other) -> bool:
        return (isinstance(other, OracleResponse)
                and self._id == other._id
                and self._response_code == other._response_code
                and self._result == other._result)


class OracleResponseCode(enum.IntEnum):
    # Indicates that the request has been successfully completed.
    Success = 0x00

    # Indicates that the protocol of the request is not supported.
    ProtocolNotSupported = 0x10

    # Indicates that the oracle nodes cannot reach a consensus on the result of the request.
    ConsensusUnreachable = 0x12

    # Indicates that the requested Uri does not exist.
    NotFound = 0x14

    # Indicates that the request was not completed within the specified time.
    Timeout = 0x16

    # Indicates that there is no permission to request the resource.
    Forbidden = 0x18

    # Indicates that the data for the response is too large.
    ResponseTooLarge = 0x1a

    # Indicates that the request failed due to insufficient balance.
    InsufficientFunds = 0x1c

    # Indicates that the request failed due to other errors.
    Error = 0xff



from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.interop.iterator import IteratorType
from boa3.model.standards.neostandard import INeoStandard
from boa3.model.standards.standardmethod import StandardMethod
from boa3.model.type.collection.sequence.uint160type import UInt160Type
from boa3.model.type.primitive.bytestringtype import ByteStringType
from boa3.model.type.type import Type


class Nep11Standard(INeoStandard):
    def __init__(self):
        type_uint160 = UInt160Type.build()
        type_iterator = IteratorType.build()
        type_bytestring = ByteStringType.build()

        methods = [
            StandardMethod('symbol', safe=True,
                           args={},
                           return_type=Type.str),
            StandardMethod('decimals', safe=True,
                           args={},
                           return_type=Type.int),
            StandardMethod('totalSupply', safe=True,
                           args={},
                           return_type=Type.int),
            StandardMethod('balanceOf', safe=True,
                           args={'owner': type_uint160},
                           return_type=Type.int),
            StandardMethod('tokensOf', safe=True,
                           args={'owner': type_uint160},
                           return_type=type_iterator),
            StandardMethod('transfer',
                           args={'to': type_uint160,
                                 'tokenId': type_bytestring,
                                 'data': Type.any},
                           return_type=Type.bool),
            StandardMethod('ownerOf', safe=True,
                           args={'tokenId': type_bytestring},
                           return_type=type_uint160),
        ]

        optionals = [
            StandardMethod('tokens', safe=True,
                           args={},
                           return_type=type_iterator),
            StandardMethod('properties', safe=True,
                           args={
                               'tokenId': type_bytestring,
                           },
                           return_type=Type.dict),
        ]

        events = [
            Builtin.Nep11Transfer
        ]

        super().__init__(methods, events, optionals)



__all__ = ['OracleResponseCodeType',
           'OracleType',
           ]

from boa3.model.builtin.interop.oracle.oracleresponsecodetype import OracleResponseCodeType
from boa3.model.builtin.interop.oracle.oracletype import OracleType



from enum import IntEnum


class ContractParameterType(IntEnum):
    Any = 0x00
    Boolean = 0x10
    Integer = 0x11
    ByteArray = 0x12
    String = 0x13
    Hash160 = 0x14
    Hash256 = 0x15
    PublicKey = 0x16
    Signature = 0x17
    Array = 0x20
    Map = 0x22
    InteropInterface = 0x30
    Void = 0xff

    @classmethod
    def _get_by_name(cls, name: str) -> int:
        try:
            value = cls.__getattr__(name)
        except BaseException:
            value = cls.Any

        return int(value)



__all__ = ['BlockType',
           'CurrentHashProperty',
           'CurrentIndexProperty',
           'GetBlockMethod',
           'GetContractMethod',
           'GetCurrentIndexMethod',
           'GetTransactionMethod',
           'GetTransactionFromBlockMethod',
           'GetTransactionHeightMethod',
           'GetTransactionSignersMethod',
           'GetTransactionVMStateMethod',
           'SignerType',
           'TransactionType',
           'VMStateType',
           'WitnessConditionEnumType',
           'WitnessConditionType',
           'WitnessRuleActionType',
           'WitnessRuleType',
           'WitnessScopeType'
           ]

from boa3.model.builtin.interop.blockchain.blocktype import BlockType
from boa3.model.builtin.interop.blockchain.currenthashmethod import CurrentHashProperty
from boa3.model.builtin.interop.blockchain.currentindexmethod import CurrentIndexProperty, GetCurrentIndexMethod
from boa3.model.builtin.interop.blockchain.getblockmethod import GetBlockMethod
from boa3.model.builtin.interop.blockchain.getcontractmethod import GetContractMethod
from boa3.model.builtin.interop.blockchain.gettransactionfromblockmethod import GetTransactionFromBlockMethod
from boa3.model.builtin.interop.blockchain.gettransactionheightmethod import GetTransactionHeightMethod
from boa3.model.builtin.interop.blockchain.gettransactionmethod import GetTransactionMethod
from boa3.model.builtin.interop.blockchain.gettransactionsignersmethod import GetTransactionSignersMethod
from boa3.model.builtin.interop.blockchain.gettransactionvmstatemethod import GetTransactionVMStateMethod
from boa3.model.builtin.interop.blockchain.signertype import SignerType
from boa3.model.builtin.interop.blockchain.transactiontype import TransactionType
from boa3.model.builtin.interop.blockchain.vmstatetype import VMStateType
from boa3.model.builtin.interop.blockchain.witnessconditionenumtype import WitnessConditionType as WitnessConditionEnumType
from boa3.model.builtin.interop.blockchain.witnessconditiontype import WitnessConditionType
from boa3.model.builtin.interop.blockchain.witnessruleactiontype import WitnessRuleActionType
from boa3.model.builtin.interop.blockchain.witnessruletype import WitnessRuleType
from boa3.model.builtin.interop.blockchain.witnessscopeenumtype import WitnessScopeType



from typing import Dict, List


class NeoExpressCommand:
    def __init__(self, command_id: str, args: List[str] = None, options: Dict[str, str] = None):
        if not isinstance(args, list):
            args = []

        if not isinstance(options, dict):
            options = {}

        self._command_id = command_id
        self._args = args
        self._options = options

    def cli_command(self) -> str:
        command = [self._command_id]

        for o in self._options.items():
            command.extend(o)

        command.extend(self._args)

        return ' '.join(command)

    def __str__(self) -> str:
        return self._command_id



from boa3.model.identifiedsymbol import IdentifiedSymbol


class __NeoMetadataType(IdentifiedSymbol):
    """
    A class used to represent the metadata object
    """

    def __init__(self):
        identifier = 'NeoMetadata'
        super().__init__(identifier)

    @property
    def shadowing_name(self) -> str:
        return 'type'

    def __str__(self) -> str:
        return self.identifier


MetadataTypeSingleton = __NeoMetadataType()



from __future__ import annotations

from enum import Enum
from typing import List


class AbiType(str, Enum):
    Signature = 'Signature'
    Boolean = 'Boolean'
    Integer = 'Integer'
    Hash160 = 'Hash160'
    Hash256 = 'Hash256'
    ByteArray = 'ByteArray'
    PublicKey = 'PublicKey'
    String = 'String'
    Array = 'Array'
    Map = 'Map'
    InteropInterface = 'InteropInterface'
    Any = 'Any'
    Void = 'Void'

    @staticmethod
    def union(abi_types: List[AbiType]) -> AbiType:
        if len(abi_types) == 0:
            return AbiType.Any

        if len(abi_types) == 1:
            return abi_types[0]

        abis = abi_types[:1]
        for abi in abi_types[1:]:
            if abi not in abis:
                abis.append(abi)

        if len(abis) == 1:
            return abis[0]

        generic_mapping = {
            AbiType.Signature: AbiType.ByteArray,
            AbiType.Boolean: AbiType.Integer,
            AbiType.Integer: AbiType.Integer,
            AbiType.Hash160: AbiType.ByteArray,
            AbiType.Hash256: AbiType.ByteArray,
            AbiType.ByteArray: AbiType.ByteArray,
            AbiType.PublicKey: AbiType.ByteArray,
            AbiType.String: AbiType.ByteArray,
            AbiType.Array: AbiType.Array,
            AbiType.Map: AbiType.Map,
            AbiType.InteropInterface: AbiType.InteropInterface,
            AbiType.Any: AbiType.Any,
            AbiType.Void: AbiType.Void
        }

        generic_abis = [generic_mapping[abis[0]]]
        for abi in abis[1:]:
            generic = generic_mapping[abi]
            if generic not in generic_abis:
                generic_abis.append(generic)

        if len(generic_abis) == 1:
            return generic_abis[0]
        else:
            return AbiType.Any



__all__ = ['ContractAbiType',
           'ContractEventDescriptorType',
           'ContractGroupType',
           'ContractGroupType',
           'ContractManifestType',
           'ContractMethodDescriptorType',
           'ContractParameterDefinitionType',
           'ContractParameterType',
           'ContractPermissionDescriptorType',
           'ContractPermissionType'
           ]

from boa3.model.builtin.interop.contract.contractmanifest.contractabitype import ContractAbiType
from boa3.model.builtin.interop.contract.contractmanifest.contracteventdescriptortype import ContractEventDescriptorType
from boa3.model.builtin.interop.contract.contractmanifest.contractgrouptype import ContractGroupType
from boa3.model.builtin.interop.contract.contractmanifest.contractmanifesttype import ContractManifestType
from boa3.model.builtin.interop.contract.contractmanifest.contractmethoddescriptortype import \
    ContractMethodDescriptorType
from boa3.model.builtin.interop.contract.contractmanifest.contractparameterdefinitiontype import \
    ContractParameterDefinitionType
from boa3.model.builtin.interop.contract.contractmanifest.contractparametertype import ContractParameterType
from boa3.model.builtin.interop.contract.contractmanifest.contractpermissiondescriptortype import \
    ContractPermissionDescriptorType
from boa3.model.builtin.interop.contract.contractmanifest.contractpermissiontype import ContractPermissionType



from os import path
from typing import Any, Dict, List, Optional, Tuple, Union

from boa3 import constants
from boa3.neo.smart_contract.VoidType import VoidType
from boa3.neo.smart_contract.notification import Notification
from boa3.neo.utils import contract_parameter_to_json, stack_item_from_json
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType
from boa3.neo.vm.type.String import String
from boa3.neo3.core.types import UInt160
from boa3.neo3.vm import VMState, vmstate
from boa3_test.tests.test_classes.block import Block
from boa3_test.tests.test_classes.contractcollection import ContractCollection
from boa3_test.tests.test_classes.signer import Signer
from boa3_test.tests.test_classes.storage import Storage
from boa3_test.tests.test_classes.testcontract import TestContract
from boa3_test.tests.test_classes.transaction import Transaction
from boa3_test.tests.test_classes.transactionattribute import oracleresponse
from boa3_test.tests.test_classes.witnessscope import WitnessScope


class TestEngine:
    def __init__(self, root_path: Optional[str] = None):
        if root_path is None:
            from boa3 import env
            root_path = env.TEST_ENGINE_DIRECTORY

        engine_path = '{0}/Neo.TestEngine.dll'.format(root_path)
        if not path.exists(engine_path):
            raise FileNotFoundError(
                "File at {0} was not found.\n"
                "Visit the docs or the README file and search for 'TestEngine' to correctly install it."
                .format(engine_path))

        self._test_engine_path = engine_path
        self._vm_state: VMState = VMState.NONE
        self._executed_script_hash: Optional[UInt160] = None
        self._gas_consumed: int = 0
        self._result_stack: List[Any] = []

        self._storage: Storage = Storage()
        self._notifications: List[Notification] = []
        self._height: int = 0
        self._blocks: List[Block] = []

        self._current_tx: Optional[Transaction] = None
        self._accounts: List[Signer] = []
        self._contract_paths: ContractCollection = ContractCollection()

        self._error_message: Optional[str] = None
        self._neo_balance_prefix: bytes = b'\x14'

    @property
    def error(self) -> Optional[str]:
        return self._error_message

    @property
    def vm_state(self) -> VMState:
        return self._vm_state

    @property
    def gas_consumed(self) -> int:
        return self._gas_consumed

    @property
    def executed_script_hash(self) -> Optional[UInt160]:
        return self._executed_script_hash

    @property
    def result_stack(self) -> List[Any]:
        return self._result_stack.copy()

    @property
    def notifications(self) -> List[Notification]:
        return self._notifications.copy()

    def get_events(self, event_name: str = None, origin: Union[UInt160, bytes] = None) -> List[Notification]:
        if origin is None:
            return [n for n in self._notifications if n.name == event_name or event_name is None]
        else:
            origin_bytes = origin.to_array() if isinstance(origin, UInt160) else bytes(origin)
            return [n for n in self._notifications if ((n.name == event_name or event_name is None)
                                                       and n.origin == origin_bytes)]

    @property
    def storage(self) -> Storage:
        return self._storage.copy()

    def storage_get(self, key: Union[str, bytes], contract_path: str) -> Any:
        if isinstance(key, str):
            key = String(key).to_bytes()

        if contract_path.endswith('.py'):
            contract_path = contract_path.replace('.py', '.nef')
        if contract_path not in self.contracts:
            return None

        contract_id = self._get_contract_id(contract_path)
        storage_key = Storage.build_key(key, contract_id)
        if storage_key in self._storage:
            return self._storage[storage_key]
        else:
            return None

    def storage_put(self, key: Union[str, bytes], value: Any, contract_path: str):
        if isinstance(key, str):
            key = String(key).to_bytes()

        if contract_path.endswith('.py'):
            contract_path = contract_path.replace('.py', '.nef')
        if contract_path in self.contracts:
            contract_id = self._get_contract_id(contract_path)
            storage_key = Storage.build_key(key, contract_id)

            self._storage[storage_key] = value

    def set_storage(self, storage: Dict[Tuple[Union[str, bytes], str], Any]):
        self._storage.clear()
        for (key, contract_path), value in storage.items():
            self.storage_put(key, value, contract_path)

    def storage_delete(self, key: Union[str, bytes], contract_path: str):
        if isinstance(key, str):
            key = String(key).to_bytes()

        if contract_path.endswith('.py'):
            contract_path = contract_path.replace('.py', '.nef')
        if contract_path not in self.contracts:
            return None

        contract_id = self._get_contract_id(contract_path)
        storage_key = Storage.build_key(key, contract_id)

        if storage_key in self._storage:
            self._storage.pop(key)

    def add_neo(self, script_hash: bytes, amount: int) -> bool:
        return self._storage.add_token(constants.NEO_SCRIPT, script_hash, amount)

    def add_gas(self, script_hash: bytes, amount: int) -> bool:
        return self._storage.add_token(constants.GAS_SCRIPT, script_hash, amount)

    def add_signer_account(self, account_address: bytes, account_scope: WitnessScope = WitnessScope.CalledByEntry):
        account = Signer(UInt160(account_address), account_scope)
        if account not in self._accounts:
            self._accounts.append(account)

    @property
    def contracts(self) -> List[str]:
        return [contract.path for contract in self._contract_paths]

    def add_contract(self, contract_nef_path: str):
        if contract_nef_path.endswith('.nef') and contract_nef_path not in self._contract_paths:
            manifest_path = contract_nef_path.replace('.nef', '.manifest.json')
            if not path.exists(manifest_path):
                raise FileNotFoundError(manifest_path)
            self._contract_paths.append(TestContract(contract_nef_path, manifest_path))

    def remove_contract(self, contract_index_or_path: Union[int, str]):
        if isinstance(contract_index_or_path, str):
            contracts = self.contracts
            if contract_index_or_path in contracts:
                index = contracts.index(contract_index_or_path)
            else:
                index = -1
        else:
            index = contract_index_or_path

        if 0 <= index < len(self._contract_paths):
            return self._contract_paths.pop(index)

    def _get_contract_id(self, contract_path: str) -> int:
        if path.isfile(contract_path):
            with open(contract_path, mode='rb') as nef:
                from boa3.neo.contracts.neffile import NefFile
                script_hash = NefFile.deserialize(nef.read()).script_hash

            return self._storage.get_contract_id(script_hash)

        contracts = self.contracts
        if contract_path in contracts:
            return contracts.index(contract_path)
        return -1

    @property
    def height(self) -> int:
        return self.current_block.index if self.current_block is not None else self._height

    @property
    def blocks(self) -> List[Block]:
        return sorted(self._blocks, key=lambda block: block.index)

    @property
    def current_block(self) -> Optional[Block]:
        return self.blocks[-1] if len(self._blocks) > 0 else None

    def increase_block(self, new_height: int = None) -> Block:
        if self.current_block is None:
            if new_height is None or new_height <= self._height:
                new_height = self._height
        else:
            if new_height is None or new_height <= self._height:
                new_height = self.height + 1
            self._height = new_height

        if new_height < 1:
            # don't use height 0 because this is the genesis block index
            new_height = 1
        new_block = Block(new_height)
        self.add_block(new_block)
        return new_block

    def add_block(self, block: Block) -> bool:
        success = len(list(filter(lambda b: b.index == block.index, self._blocks))) == 0
        if success:
            self._blocks.append(block)
        return success

    def get_transactions(self) -> List[Transaction]:
        txs = []
        for block in self._blocks:
            txs.extend(block.get_transactions())
        return txs

    def add_transaction(self, *transaction: Transaction):
        if self.current_block is None:
            self.increase_block()

        current_block = self.current_block
        for tx in transaction:
            current_block.add_transaction(tx)

    def run_oracle_response(self, request_id: int, oracle_response: oracleresponse.OracleResponseCode,
                            result: bytes, reset_engine: bool = False,
                            rollback_on_fault: bool = True) -> Any:
        request_ids = [x.arguments[0] if isinstance(x.arguments, (tuple, list)) and len(x.arguments) > 0
                       else x.arguments
                       for x in self.get_events('OracleRequest', constants.ORACLE_SCRIPT)]

        assert request_id in request_ids, 'Request ID not found'
        self._current_tx = Transaction(b'')
        self._current_tx.add_attribute(oracleresponse.OracleResponse(request_id, oracle_response, result))

        return self.run(UInt160(constants.ORACLE_SCRIPT), 'finish',
                        reset_engine=reset_engine,
                        rollback_on_fault=rollback_on_fault)

    def run(self, contract_id: Union[str, UInt160], method: str, *arguments: Any, reset_engine: bool = False,
            rollback_on_fault: bool = True) -> Any:
        import json
        import subprocess

        if isinstance(contract_id, str) and contract_id not in self.contracts:
            self.add_contract(contract_id)

        # build an UInt160 value if the contract_id is not a path
        if isinstance(contract_id, bytes) and not isinstance(contract_id, UInt160):
            contract_id = UInt160(contract_id)

        test_engine_args = self.to_json(contract_id, method, *arguments)
        param_json = json.dumps(test_engine_args, separators=(',', ':'))

        try:
            process = subprocess.Popen(['dotnet', self._test_engine_path, param_json],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       text=True)
        except BaseException:
            json_path = '{0}/test-engine-test.json'.format(path.curdir)
            with open(json_path, 'wb+') as json_file:
                json_file.write(String(param_json).to_bytes())
                json_file.close()

            process = subprocess.Popen(['dotnet', self._test_engine_path, json_path],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       text=True)

        stdout, stderr = process.communicate()

        if reset_engine:
            self.reset_engine()
        else:
            self.reset_state()

        stdout = stdout.splitlines()[-1]

        try:
            result = json.loads(stdout)

            self._error_message = result['error'] if 'error' in result else None

            if 'vmstate' in result:
                self._vm_state = vmstate.get_vm_state(result['vmstate'])

            if 'executedscripthash' in result:
                self._executed_script_hash = UInt160.from_string(result['executedscripthash'])

            if 'gasconsumed' in result:
                self._gas_consumed = int(result['gasconsumed'])

            if 'resultstack' in result:
                if isinstance(result['resultstack'], list):
                    self._result_stack = [stack_item_from_json(value) for value in result['resultstack']]
                else:
                    self._result_stack = [stack_item_from_json(result['resultstack'])]

            if self._vm_state is VMState.HALT or not rollback_on_fault:
                if 'notifications' in result:
                    json_storage = result['notifications']
                    if not isinstance(json_storage, list):
                        json_storage = [json_storage]

                    notifications = []
                    for n in json_storage:
                        new = Notification.from_json(n)
                        if new is not None:
                            notifications.append(new)
                    self._notifications.extend(notifications)

                if 'storage' in result:
                    json_storage = result['storage']
                    self._storage = Storage.from_json(json_storage)

                    for contract in self._contract_paths.copy():
                        if (not isinstance(contract, TestContract)
                                or contract.script_hash is None
                                or not self._storage.has_contract(contract.script_hash)):
                            self.remove_contract(contract.path)

                if 'currentblock' in result:
                    current_block = Block.from_json(result['currentblock'])

                    existing_block = next((block for block in self._blocks if block.index == current_block.index), None)
                    if existing_block is not None:
                        self._blocks.remove(existing_block)
                    self._blocks.append(current_block)

                if 'transaction' in result and self._vm_state is VMState.HALT:
                    block = self.current_block
                    if block is None:
                        block = self.increase_block(self.height)

                    tx = Transaction.from_json(result['transaction'])
                    block.add_transaction(tx)

        except BaseException as e:
            self._error_message = str(e)

        if len(self._result_stack) > 0:
            return self._filter_result(self._result_stack[0], method, contract_id)
        return VoidType

    def reset_state(self):
        self._vm_state = VMState.NONE
        self._executed_script_hash = None
        self._gas_consumed = 0
        self._result_stack = []
        self._accounts = []
        self._current_tx = None
        self._error_message = None

    def reset_engine(self):
        self.reset_state()
        self._notifications.clear()
        self._storage.clear()

    def _filter_result(self, value: Any, called_method: str, contract_id=None):
        if contract_id is None:
            contract_id = self._executed_script_hash

        called_contract = self._contract_paths[contract_id]
        if not isinstance(called_contract, TestContract):
            return value

        converted_value = value
        expected_return_type = called_contract.get_method_return_type(called_method)

        if expected_return_type is StackItemType.Boolean:
            converted_value = bool(value)
        elif expected_return_type is StackItemType.Integer:
            if isinstance(value, str):
                converted_value = String(value).to_bytes(min_length=1)

            if isinstance(converted_value, bytes):
                converted_value = Integer.from_bytes(value)
            else:
                try:
                    converted_value = int(value)
                except BaseException:
                    converted_value = value
        elif expected_return_type is StackItemType.ByteString:
            if isinstance(value, int):
                converted_value = Integer(value).to_byte_array(min_length=1)

        return converted_value

    def to_json(self, contract_id: Union[str, UInt160], method: str, *args: Any) -> Dict[str, Any]:
        json = {
            'path': contract_id if isinstance(contract_id, str) else '',
            'scripthash': str(contract_id) if not isinstance(contract_id, str) else None,
            'method': method,
            'arguments': [contract_parameter_to_json(x) for x in args],
            'storage': self._storage.to_json(),
            'contracts': [{'nef': contract_path} for contract_path in self.contracts],
            'signeraccounts': [address.to_json() for address in self._accounts],
            'height': self.height,
            'blocks': [block.to_json() for block in self.blocks]
        }
        if isinstance(self._current_tx, Transaction):
            json['currenttx'] = self._current_tx.to_json()
        return json



from typing import Any, Dict, Iterable, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ReversedMethod(IBuiltinMethod):
    def __init__(self, args_type: SequenceType = None):
        from boa3.model.type.type import Type
        identifier = 'reversed'
        if not isinstance(args_type, SequenceType):
            args_type = Type.sequence

        args: Dict[str, Variable] = {'sequence': Variable(args_type)}

        super().__init__(identifier, args, return_type=Type.reversed)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self.args['sequence'].type is Type.sequence:
            return self.raw_identifier
        return '-{0}_from_{1}'.format(self._identifier, self.args['sequence'].type.identifier)

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Iterable) and len(value) > 0:
            value = value[0]

        if not isinstance(value, SequenceType):
            return ReversedMethod()

        if type(value) == type(self.args['sequence'].type):
            return self

        return ReversedMethod(value)

    @property
    def _body(self) -> Optional[str]:
        return

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.type import Type
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')

        if_statement = [
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, Type.str.stack_item),  # if isinstance(arg1, (str, bytes)) continue
            # if not isinstance(arg1, (str, bytes)) will jump into the else_body
        ]

        if_true_body = [    # if isinstance(arg1, (str, bytes)) body will continue until else_body
            (Opcode.NEWARRAY0, b''),  # list_aux = []
            (Opcode.SWAP, b''),
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),       # limit = len(arg)
            (Opcode.PUSH0, b''),      # index = 0
        ]

        while_statement = [           # index and limit will be at the top of the stack
            (Opcode.DUP, b''),        # verifies if index < limit
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.LT, b''),
            jmp_place_holder          # if index >= limit, go to break_opcode
        ]

        while_body = [                # will add every byte or character into the array and increase the index
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT,
             Type.int.stack_item if Type.bytes.is_type_of(self.args['sequence'].type) else Type.str.stack_item),
            (Opcode.APPEND, b''),    # list_aux.append(arg[index])
            (Opcode.INC, b''),       # index++
            # returns to beginning of the while to verify if it index < limit
        ]

        jmp_back_to_while_statement = Opcode.get_jump_and_data(Opcode.JMP, -get_bytes_count(while_body + while_statement))
        while_body.append(jmp_back_to_while_statement)

        jmp_to_break_body = Opcode.get_jump_and_data(Opcode.JMPIFNOT, get_bytes_count(while_body), True)
        while_statement[-1] = jmp_to_break_body

        break_opcode = [        # remove auxiliary values from stack after leaving `while`
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            # skips the else_body, because it's a bytestring value and unpacking and packing are expensive
        ]

        else_body = [    # creates a new array from an array, used when the arg is not a str or bytes value
            (Opcode.UNPACK, b''),
            (Opcode.PACK, b''),
        ]

        jmp_else_body = Opcode.get_jump_and_data(Opcode.JMP, get_bytes_count(else_body), True)
        break_opcode.append(jmp_else_body)

        if_true_body = if_true_body + while_statement + while_body + break_opcode

        jmp_if_true_body = Opcode.get_jump_and_data(Opcode.JMPIFNOT, get_bytes_count(if_true_body))
        if_statement.append(jmp_if_true_body)

        reverse_array = [   # reverse the array
            (Opcode.DUP, b''),
            (Opcode.REVERSEITEMS, b''),     # return reversed(arg)
        ]

        return (
            if_statement +
            if_true_body +
            else_body +
            reverse_array
        )



import ast
from typing import Dict, List

from boa3.model.builtin.interop.nativecontract.Oracle.getoraclescripthashmethod import OracleContract
from boa3.model.builtin.interop.nativecontract.nativecontractmethod import NativeContractMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class OracleMethod(NativeContractMethod):

    def __init__(self, identifier: str, syscall: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(OracleContract.getter, identifier, syscall, args, defaults, return_type)



from typing import Any


def json_serialize(item: Any) -> str:
    """
    Serializes an item into a json.

    :param item: The item that will be serialized
    :type item: Any
    :return: The serialized item
    :rtype: str

    :raise Exception: raised if the item is an integer value out of the Neo's accepted range, is a dictionary with a
        bytearray key, or isn't serializable.
    """
    pass


def json_deserialize(json: str) -> Any:
    """
    Deserializes a json into some valid type.

    :param json: A json that will be deserialized
    :type json: str
    :return: The deserialized json
    :rtype: Any

    :raise Exception: raised if jsons deserialization is not valid.
    """
    pass



from typing import List, Optional, Tuple

from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.type.collection.icollection import ICollectionType
from boa3.model.type.collection.mapping.mappingtype import MappingType
from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class CollectionNotMembership(BinaryOperation):
    """
    A class used to represent a collection not membership operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.collection]

    def __init__(self, left: IType = Type.any, right: IType = None):
        self.operator: Operator = Operator.NotIn

        if not isinstance(right, ICollectionType):
            right = Type.collection

        if not right.item_type.is_type_of(left) or right.item_type is Type.any and left is not Type.any:
            right = Type.collection.build(Type.sequence.build(left))

        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        if not isinstance(right, ICollectionType):
            return False

        if isinstance(right, PrimitiveType) and (right.is_type_of(left) or left.is_type_of(right)):
            return True

        if isinstance(right, MappingType):
            reference_type = right.key_type
        else:
            reference_type = right.item_type

        return reference_type.is_type_of(left)

    def _get_result(self, left: IType, right: IType) -> IType:
        return Type.bool

    def get_valid_operand_for_validation(self, left_operand: IType,
                                         right_operand: IType = None) -> Tuple[Optional[IType], Optional[IType]]:
        if isinstance(right_operand, ICollectionType):
            left = right_operand.item_type if not isinstance(right_operand, MappingType) else right_operand.key_type
            return left, right_operand

        return left_operand, self.right_type

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.operation.binary.additional import CollectionMembership
        return CollectionMembership().opcode + [(Opcode.NOT, b'')]



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.classmethod.popmethod import PopMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class PopDictDefaultMethod(PopMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type

        if not Type.dict.is_type_of(arg_value):
            arg_value = Type.dict

        args: Dict[str, Variable] = {
            'self': Variable(arg_value),
            'key': Variable(arg_value.valid_key),
            'default': Variable(Type.any)
        }

        super().__init__(args, return_type=arg_value.value_type)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')

        put_default_at_bottom = [
            (Opcode.REVERSE3, b''),
            (Opcode.SWAP, b''),
            (Opcode.REVERSE3, b''),
        ]

        prepare_values = [
            (Opcode.DUP, b''),
            (Opcode.SIGN, b''),
            (Opcode.PUSHM1, b''),
            (Opcode.JMPNE, Integer(5).to_byte_array(min_length=1, signed=True)),
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
        ]

        verify_has_key = [
            (Opcode.HASKEY, b''),
            jmp_place_holder
        ]

        pick_from_dict = [
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.SWAP, b''),
        ]

        pop_from_dict = [
            (Opcode.REMOVE, b''),
            (Opcode.NIP, b''),
            jmp_place_holder
        ]

        return_default = [
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        num_jmp_code = get_bytes_count(return_default)
        pop_from_dict[-1] = Opcode.get_jump_and_data(Opcode.JMP, num_jmp_code, True)

        num_jmp_code = get_bytes_count(pick_from_dict + pop_from_dict)
        verify_has_key[-1] = Opcode.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)

        return (
            put_default_at_bottom +
            prepare_values +
            verify_has_key +
            pick_from_dict +
            pop_from_dict +
            return_default
        )



import ast
import importlib.util
import os
import sys
from typing import Dict, List, Optional

from boa3 import constants
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.model import imports
from boa3.model.symbol import ISymbol
from boa3.model.type.type import Type


class ImportAnalyser(IAstAnalyser):

    def __init__(self, import_target: str, root_folder: str,
                 importer_file: Optional[str] = None,
                 import_stack: List[str] = None,
                 already_imported_modules: dict = None,
                 log: bool = False,
                 get_entry: bool = False):

        self.can_be_imported: bool = False
        self.is_builtin_import: bool = False
        self.is_namespace_package: bool = False
        self.recursive_import: bool = False
        self._import_identifier: str = import_target
        self._get_from_entry: bool = get_entry and (os.path.isfile(import_target) or os.path.isdir(import_target))

        from boa3.analyser.analyser import Analyser
        self._imported_files: Dict[str, Analyser] = (already_imported_modules
                                                     if isinstance(already_imported_modules, dict)
                                                     else {})
        self._import_stack: List[str] = import_stack if isinstance(import_stack, list) else []
        self.analyser: Analyser = None  # set if the import is successful
        self._submodule_search_locations = None

        if isinstance(root_folder, str):
            if os.path.isfile(root_folder):
                root = os.path.dirname(root_folder)
            elif os.path.isdir(root_folder):
                root = root_folder
            else:
                root = os.path.dirname(importer_file)
        else:
            root = os.path.dirname(importer_file)

        super().__init__(ast.Module(body=[]), root_folder=root, log=log)

        if self._get_from_entry:
            self.path: str = import_target.replace(os.sep, constants.PATH_SEPARATOR)
            self._find_package(import_target, importer_file)
            return

        sys.path.append(self.root_folder)
        try:
            import_spec = importlib.util.find_spec(import_target)
            module_origin: str = import_spec.origin
        except BaseException:
            return
        finally:
            sys.path.remove(self.root_folder)

        self._importer_file = importer_file
        is_importing_a_module = module_origin is not None

        if not is_importing_a_module:
            self.is_namespace_package = True
            if import_spec.submodule_search_locations is None:
                return

            # submodule_search_locations type doesn't support indexing, so we need to iterate to get the first
            for module_ in import_spec.submodule_search_locations:
                module_origin: str = module_  # + "\\" + names[0] + '.py'
                break

            # if the origin is still not set, there's nothing to import
            if module_origin is None:
                return

        path: List[str] = module_origin.split(os.sep)
        self.filename = path[-1]
        self._submodule_search_locations = import_spec.submodule_search_locations
        self.path: str = module_origin.replace(os.sep, constants.PATH_SEPARATOR)

        self._find_package(module_origin, importer_file)

    @property
    def tree(self) -> ast.AST:
        return self._tree

    def export_symbols(self, identifiers: List[str] = None) -> Dict[str, ISymbol]:
        """
        Gets a dictionary that maps each exported symbol with its identifier

        :param identifiers: list of identifiers of the imported symbols
        :return:
        """
        if identifiers is None:
            identifiers = list([name for name, symbol in self.symbols.items() if symbol is not None])
        if not self.can_be_imported or not isinstance(identifiers, (list, str)):
            return {}

        if isinstance(identifiers, str):
            identifiers = [identifiers]

        if constants.IMPORT_WILDCARD in identifiers:
            symbols = self.symbols.copy()
        else:
            symbols = {symbol_id: symbol for symbol_id, symbol in self.symbols.items()
                       if symbol_id in identifiers and symbol is not None}
        return symbols

    def update_external_analysed_files(self, external_files: dict):
        if self.can_be_imported and self.path in self._imported_files and self.path not in external_files:
            external_files[self.path] = self._imported_files[self.path]

        for path, analyser in self._imported_files.items():
            if path not in external_files and analyser.is_analysed:
                external_files[path] = analyser

    def _find_package(self, module_origin: str, origin_file: Optional[str] = None):
        path: List[str] = module_origin.split(os.sep)

        package = imports.builtin.get_package(self._import_identifier)
        if hasattr(package, 'symbols'):
            if hasattr(package, 'inner_packages'):
                # when have symbol and packages with the same id, prioritize symbol
                self.symbols: Dict[str, ISymbol] = package.inner_packages
                self.symbols.update(package.symbols)
            else:
                self.symbols = package.symbols

            self.can_be_imported = True
            self.is_builtin_import = True
            return

        if (os.path.commonpath([self.path, constants.BOA_PACKAGE_PATH]) != constants.BOA_PACKAGE_PATH or
                ('boa3' in path and constants.PATH_SEPARATOR.join(path[path.index('boa3'):]).startswith('boa3/builtin'))):
            # doesn't analyse boa3.builtin packages that aren't included in the imports.builtin as an user module
            import re

            inside_python_folder = any(re.search(r'python(\d\.?)*', folder.lower()) for folder in path)
            updated_tree = None

            if not (inside_python_folder and 'lib' in path):
                # check circular imports to avoid recursions inside the compiler
                if self.path in self._import_stack:
                    self.recursive_import = True
                    return

                # TODO: only user modules and typing lib imports are implemented
                try:
                    if self.path in self._imported_files:
                        analyser = self._imported_files[self.path]
                    else:
                        from boa3.analyser.analyser import Analyser
                        origin = origin_file.replace(os.sep, constants.PATH_SEPARATOR)
                        files = self._import_stack
                        files.append(origin)
                        if self.is_namespace_package:
                            analyser = Analyser(self.tree, module_origin, self.root_folder, self._log)
                            if self._include_inner_packages(analyser):
                                analyser.is_analysed = True
                                self._imported_files[self.path] = analyser
                        else:
                            analyser = Analyser.analyse(module_origin, root=self.root_folder,
                                                        imported_files=self._imported_files,
                                                        import_stack=files, log=self._log)
                            self._include_inner_packages(analyser)

                        if analyser.is_analysed:
                            self._imported_files[self.path] = analyser

                    # include only imported symbols
                    if analyser.is_analysed:
                        for symbol_id, symbol in analyser.symbol_table.items():
                            if symbol_id not in Type.all_types():
                                if not self._get_from_entry:
                                    symbol.defined_by_entry = False
                                self.symbols[symbol_id] = symbol

                    self.errors.extend(analyser.errors)
                    self.warnings.extend(analyser.warnings)

                    updated_tree = analyser.ast_tree
                    self.analyser = analyser
                    self.can_be_imported = analyser.is_analysed
                except FileNotFoundError:
                    self.can_be_imported = False

                if updated_tree is not None:
                    self._tree = updated_tree

    def _include_inner_packages(self, analyser) -> bool:
        if self.path.endswith('.py') and self.filename != f'{constants.INIT_METHOD_ID}.py':
            return False

        import pkgutil
        from boa3.model.imports.importsymbol import Import
        from boa3.model.imports.package import Package

        inner_packages_have_errors = False
        modules = {}
        for importer, modname, is_pkg in pkgutil.iter_modules(self._submodule_search_locations):
            mod_target = self._import_identifier + constants.ATTRIBUTE_NAME_SEPARATOR + modname
            import_analyser = ImportAnalyser(mod_target, self.root_folder,
                                             importer_file=self._importer_file,
                                             import_stack=self._import_stack,
                                             already_imported_modules=self._imported_files,
                                             log=self._log)

            if not inner_packages_have_errors and not import_analyser.can_be_imported:
                inner_packages_have_errors = True

            imported = Package(identifier=modname,
                               other_symbols=import_analyser.symbols,
                               import_origin=Import(import_analyser.path,
                                                    import_analyser._tree,
                                                    import_analyser))
            modules[modname] = imported

        if len(modules) > 0 and hasattr(analyser, 'symbol_table') and isinstance(analyser.symbol_table, dict):
            analyser.symbol_table.update(modules)

        return not inner_packages_have_errors



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ExtendMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        from boa3.model.type.type import Type
        if not isinstance(sequence_type, MutableSequenceType):
            sequence_type = Type.mutableSequence

        self_arg = Variable(sequence_type)
        item_arg = Variable(Type.sequence.build_collection(sequence_type.value_type))

        identifier = 'extend'
        args: Dict[str, Variable] = {'self': self_arg, 'item': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        sequence_type: IType = params[0].type
        iterator_type: IType = params[1].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        if not isinstance(iterator_type, SequenceType):
            return False
        return sequence_type.value_type.is_type_of(iterator_type.value_type)

    @property
    def stores_on_slot(self) -> bool:
        return True

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.StackItem import StackItemType
        return [
            (Opcode.OVER, b''),
            (Opcode.ISTYPE, StackItemType.Array),
            (Opcode.JMPIF, Integer(5).to_byte_array(signed=True, min_length=1)),
            (Opcode.CAT, b''),
            (Opcode.JMP, Integer(18).to_byte_array(signed=True, min_length=1)),
            (Opcode.UNPACK, b''),       # get the values, top of stack will be the array size
            (Opcode.JMP, Integer(9).to_byte_array(signed=True, min_length=1)),  # begin while
            (Opcode.DUP, b''),
            (Opcode.INC, b''),          # push the array to the top of the stack
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.ROLL, b''),         # get the first value that wasn't appended yet
            (Opcode.APPEND, b''),       # append the value to the array
            (Opcode.DEC, b''),
            (Opcode.DUP, b''),          # when the array is empty, stop the loop
            (Opcode.JMPIF, Integer(-8).to_byte_array(signed=True, min_length=1)),
            (Opcode.DROP, b''),  # the top items in the stack will be the extended array and the other array size (zero)
            (Opcode.DROP, b'')
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return ExtendMethod(value)
        return super().build(value)



from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.builtin.classmethod import *
from boa3.model.builtin.contract import *
from boa3.model.builtin.decorator import *
from boa3.model.builtin.internal.innerdeploymethod import InnerDeployMethod
from boa3.model.builtin.interop.interop import Interop
from boa3.model.builtin.math import *
from boa3.model.builtin.method import *
from boa3.model.builtin.neometadatatype import MetadataTypeSingleton as NeoMetadataType
from boa3.model.callable import Callable
from boa3.model.event import Event as EventSymbol
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.imports.package import Package
from boa3.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.model.type.collection.sequence.uint160type import UInt160Type
from boa3.model.type.collection.sequence.uint256type import UInt256Type
from boa3.model.type.itype import IType
from boa3.model.type.math import Math
from boa3.model.type.primitive.bytestringtype import ByteStringType


class BoaPackage(str, Enum):
    Contract = 'contract'
    Interop = 'interop'
    Type = 'type'


class Builtin:
    @classmethod
    def get_symbol(cls, symbol_id: str) -> Optional[Callable]:
        for method in cls._python_builtins:
            if isinstance(method, IBuiltinCallable) and method.identifier == symbol_id:
                return method

    @classmethod
    def get_by_self(cls, symbol_id: str, self_type: IType) -> Optional[Callable]:
        for name, method in vars(cls).items():
            if (isinstance(method, IBuiltinMethod)
                    and method.identifier == symbol_id
                    and method.validate_self(self_type)):
                return method

    # builtin method
    Abs = AbsMethod()
    Exit = ExitMethod()
    IsInstance = IsInstanceMethod()
    Len = LenMethod()
    NewEvent = CreateEventMethod()
    Max = MaxIntMethod()
    Min = MinIntMethod()
    Print = PrintMethod()
    ScriptHash = ScriptHashMethod()
    StrSplit = StrSplitMethod()
    Sum = SumMethod()

    # python builtin class constructor
    Bool = BoolMethod()
    ByteArray = ByteArrayMethod()
    ByteArrayEncoding = ByteArrayEncodingMethod()
    Exception = ExceptionMethod()
    IntByteString = IntByteStringMethod()
    IntInt = IntIntMethod()
    ListBytesString = ListBytesStringMethod()
    ListGeneric = ListGenericMethod()
    ListMapping = ListMappingMethod()
    ListSequence = ListSequenceMethod()
    Range = RangeMethod()
    Reversed = ReversedMethod()
    Super = SuperMethod()

    # python class method
    BytesStringIndex = IndexBytesStringMethod()
    BytesStringIsDigit = IsDigitMethod()
    BytesStringJoin = JoinMethod()
    BytesStringLower = LowerMethod()
    BytesStringStartswith = StartsWithMethod()
    BytesStringStrip = StripMethod()
    BytesStringUpper = UpperMethod()
    CountSequenceGeneric = CountSequenceGenericMethod()
    CountSequencePrimitive = CountSequencePrimitiveMethod()
    CountStr = CountStrMethod()
    Copy = CopyListMethod()
    SequenceAppend = AppendMethod()
    SequenceClear = ClearMethod()
    SequenceExtend = ExtendMethod()
    SequenceIndex = IndexSequenceMethod()
    SequenceInsert = InsertMethod()
    SequencePop = PopSequenceMethod()
    SequenceRemove = RemoveMethod()
    SequenceReverse = ReverseMethod()
    DictKeys = MapKeysMethod()
    DictPop = PopDictMethod()
    DictPopDefault = PopDictDefaultMethod()
    DictValues = MapValuesMethod()

    # custom class methods
    ConvertToBytes = ToBytesMethod
    ConvertToInt = ToIntMethod
    ConvertToStr = ToStrMethod
    ConvertToBool = ToBoolMethod

    # builtin decorator
    ClassMethodDecorator = ClassMethodDecorator()
    InstanceMethodDecorator = InstanceMethodDecorator()
    PropertyDecorator = PropertyDecorator()
    StaticMethodDecorator = StaticMethodDecorator()

    _python_builtins: List[IdentifiedSymbol] = [Abs,
                                                ByteArray,
                                                ByteArrayEncoding,
                                                BytesStringIndex,
                                                BytesStringIsDigit,
                                                BytesStringJoin,
                                                BytesStringLower,
                                                BytesStringStartswith,
                                                BytesStringStrip,
                                                BytesStringUpper,
                                                ClassMethodDecorator,
                                                ConvertToBool,
                                                ConvertToBytes,
                                                ConvertToInt,
                                                ConvertToStr,
                                                Copy,
                                                CountSequenceGeneric,
                                                CountSequencePrimitive,
                                                CountStr,
                                                DictKeys,
                                                DictValues,
                                                Exception,
                                                Exit,
                                                IsInstance,
                                                Len,
                                                Max,
                                                Min,
                                                Print,
                                                PropertyDecorator,
                                                Range,
                                                Reversed,
                                                ScriptHash,
                                                SequenceAppend,
                                                SequenceClear,
                                                SequenceExtend,
                                                SequenceIndex,
                                                SequenceInsert,
                                                SequencePop,
                                                SequenceRemove,
                                                SequenceReverse,
                                                StaticMethodDecorator,
                                                StrSplit,
                                                Sum,
                                                Super,
                                                ]

    @classmethod
    def interop_symbols(cls, package: str = None) -> Dict[str, IdentifiedSymbol]:
        return {symbol.raw_identifier if hasattr(symbol, 'raw_identifier') else symbol.identifier: symbol
                for symbol in Interop.interop_symbols(package)}

    # boa builtin decorator
    ContractInterface = ContractDecorator()
    ContractMethodDisplayName = DisplayNameDecorator()
    Metadata = MetadataDecorator()
    Public = PublicDecorator()

    # boa builtin type
    ByteString = ByteStringType.build()
    Event = EventType
    UInt160 = UInt160Type.build()
    UInt256 = UInt256Type.build()
    ECPoint = ECPointType.build()
    NeoAccountState = NeoAccountStateType.build()

    # boa events
    Nep5Transfer = Nep5TransferEvent()
    Nep11Transfer = Nep11TransferEvent()
    Nep17Transfer = Nep17TransferEvent()

    # boa smart contract methods
    Abort = AbortMethod()

    # region boa builtin modules

    BuiltinMathCeil = DecimalCeilingMethod()
    BuiltinMathFloor = DecimalFloorMethod()

    MathModule = Package(identifier='math',
                         methods=[Math.Sqrt,
                                  BuiltinMathCeil,
                                  BuiltinMathFloor])

    _modules = [MathModule]

    # endregion

    boa_builtins: List[IdentifiedSymbol] = [ContractInterface,
                                            ContractMethodDisplayName,
                                            Event,
                                            Metadata,
                                            NeoMetadataType,
                                            NewEvent,
                                            Public,
                                            ScriptHash
                                            ] + _modules

    metadata_fields: Dict[str, Union[type, Tuple[type]]] = {
        'name': str,
        'supported_standards': list,
        'trusts': list,
        'author': (str, type(None)),
        'email': (str, type(None)),
        'description': (str, type(None)),
        'extras': dict
    }

    @classmethod
    def boa_symbols(cls) -> Dict[str, IdentifiedSymbol]:
        return {symbol.identifier: symbol for symbol in cls.boa_builtins}

    @classmethod
    def package_symbols(cls, package: str = None) -> Dict[str, IdentifiedSymbol]:
        if package in BoaPackage.__members__.values():
            return {symbol.identifier: symbol for symbol in cls._boa_symbols[package]}

        return cls.boa_symbols()

    @classmethod
    def builtin_events(cls) -> List[EventSymbol]:
        lst: List[EventSymbol] = [event for event in cls.boa_builtins if isinstance(event, EventSymbol)]

        for symbols in cls._boa_symbols.values():
            lst.extend([event for event in symbols if isinstance(event, EventSymbol)])

        return lst

    _boa_symbols: Dict[BoaPackage, List[IdentifiedSymbol]] = {
        BoaPackage.Contract: [Abort,
                              NeoAccountState,
                              Nep11Transfer,
                              Nep17Transfer,
                              Nep5Transfer,
                              ],
        BoaPackage.Interop: Interop.package_symbols,
        BoaPackage.Type: [ByteString,
                          ECPoint,
                          UInt160,
                          UInt256
                          ]
    }

    _internal_methods = [InnerDeployMethod.instance()
                         ]
    internal_methods = {method.raw_identifier: method for method in _internal_methods}



from enum import IntFlag


class TriggerType(IntFlag):
    """
    Represents the triggers for running smart contracts. Triggers enable the contract to execute different logic under
    different usage scenarios.
    """

    ON_PERSIST = 0x01
    """
    Indicate that the contract is triggered by the system to execute the OnPersist method of the native contracts.

    :meta hide-value:
    """

    POST_PERSIST = 0x02
    """
    Indicate that the contract is triggered by the system to execute the PostPersist method of the native contracts.

    :meta hide-value:
    """

    SYSTEM = ON_PERSIST | POST_PERSIST
    """
    The combination of all system triggers.

    :meta hide-value:
    """

    VERIFICATION = 0x20
    """
    Indicates that the contract is triggered by the verification of a IVerifiable.

    :meta hide-value:
    """

    APPLICATION = 0x40
    """
    Indicates that the contract is triggered by the execution of transactions.

    :meta hide-value:
    """

    ALL = ON_PERSIST | POST_PERSIST | VERIFICATION | APPLICATION
    """
    The combination of all triggers.

    :meta hide-value:
    """


class CallFlags(IntFlag):
    """
    Defines special behaviors allowed when invoking smart contracts, e.g., chain calls, sending notifications and
    modifying states.
    """

    NONE = 0
    """
    Special behaviors of the invoked contract are not allowed, such as chain calls, sending notifications, modifying 
    state, etc.

    :meta hide-value:
    """

    READ_STATES = 0b00000001
    """
    Indicates that the called contract is allowed to read states.

    :meta hide-value:
    """

    WRITE_STATES = 0b00000010
    """
    Indicates that the called contract is allowed to write states.

    :meta hide-value:
    """

    ALLOW_CALL = 0b00000100
    """
    Indicates that the called contract is allowed to call another contract.

    :meta hide-value:
    """

    ALLOW_NOTIFY = 0b00001000
    """
    Indicates that the called contract is allowed to send notifications.

    :meta hide-value:
    """

    STATES = READ_STATES | WRITE_STATES
    """
    Indicates that the called contract is allowed to read or write states.

    :meta hide-value:
    """

    READ_ONLY = READ_STATES | ALLOW_CALL
    """
    Indicates that the called contract is allowed to read states or call another contract.

    :meta hide-value:
    """

    ALL = STATES | ALLOW_CALL | ALLOW_NOTIFY
    """
    All behaviors of the invoked contract are allowed.

    :meta hide-value:
    """



import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.model.callable import Callable
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class IBuiltinCallable(Callable, IdentifiedSymbol, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(args, vararg, kwargs, defaults, return_type)
        self._identifier = identifier
        self.defined_by_entry = False  # every builtin symbol must have this variable set as False

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        """
        Gets the opcode for the method.

        :return: the opcode and its data if exists. None otherwise.
        """
        return []

    @property
    def identifier(self) -> str:
        return self._identifier



from typing import Any, Dict

from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.model.type.primitive.inttype import IntType


class FindOptionsType(IntType):
    """
    A class used to represent Neo interop FindOptions type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'FindOptions'

    @property
    def default_value(self) -> Any:
        from boa3.builtin.interop.storage import FindOptions
        return FindOptions.NONE

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if cls._is_type_of(value) or value is None:
            from boa3.model.builtin.interop.interop import Interop
            return Interop.FindOptionsType

    @classmethod
    def _is_type_of(cls, value: Any):
        from boa3.builtin.interop.storage import FindOptions
        return isinstance(value, (FindOptions, FindOptionsType))

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the module with its name
        """
        from boa3.builtin.interop.storage import FindOptions
        from boa3.model.variable import Variable

        _symbols = super().symbols
        _symbols.update({name: Variable(self) for name in FindOptions.__members__.keys()})

        return _symbols

    def get_value(self, symbol_id) -> Any:
        """
        Gets the literal value of a symbol

        :return: the value if this type has this symbol. None otherwise.
        """
        from boa3.builtin.interop.storage import FindOptions

        if symbol_id in self.symbols and symbol_id in FindOptions.__members__:
            return FindOptions.__members__[symbol_id]

        return None



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class DecimalCeilingMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'ceil'
        args: Dict[str, Variable] = {'x': Variable(Type.int),
                                     'decimals': Variable(Type.int)}
        super().__init__(identifier, args, return_type=Type.int)

    @property
    def exception_message(self) -> str:
        return "decimals cannot be negative"

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.String import String

        message = String(self.exception_message).to_bytes()

        if_negative_decimal = [
            (Opcode.PUSHDATA1, Integer(len(message)).to_byte_array(signed=True, min_length=1) + message),
            (Opcode.THROW, b''),
        ]

        decimals_unit = [
            (Opcode.OVER, b''),
            (Opcode.PUSH0, b''),
            Opcode.get_jump_and_data(Opcode.JMPGE, get_bytes_count(if_negative_decimal), True),
        ] + if_negative_decimal + [
            (Opcode.DUP, b''),
            (Opcode.REVERSE3, b''),
            Opcode.get_push_and_data(10),
            (Opcode.SWAP, b''),
            (Opcode.POW, b''),
            (Opcode.SWAP, b''),
            (Opcode.NEGATE, b''),
            (Opcode.OVER, b''),
            (Opcode.MOD, b'')
        ]

        if_negative_mod = [
            (Opcode.ADD, b''),
            (Opcode.JMP, b'')
        ]
        else_negative_mod = [
            (Opcode.NIP, b'')
        ]
        num_jmp_code = get_bytes_count(else_negative_mod)
        if_negative_mod[-1] = Opcode.get_jump_and_data(Opcode.JMP, num_jmp_code, True)

        num_jmp_code = get_bytes_count(if_negative_mod)
        floor_computation = [
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            Opcode.get_jump_and_data(Opcode.JMPGE, num_jmp_code, True),
        ] + if_negative_mod + else_negative_mod + [
            (Opcode.ADD, b'')
        ]

        return decimals_unit + floor_computation

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



import ast
from typing import Any, Dict, List, Sized

from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.variable import Variable
from boa3.neo3.core.types import UInt160


class ContractDecorator(IBuiltinDecorator):
    def __init__(self):
        from boa3.model.type.primitive.bytestringtype import ByteStringType

        identifier = 'contract'
        args: Dict[str, Variable] = {'script_hash': Variable(ByteStringType.build())}
        super().__init__(identifier, args)
        self.contract_hash = UInt160()

    def build(self, *args) -> IBuiltinDecorator:
        if isinstance(args, Sized) and len(args) > 0 and isinstance(args[0], ast.AST):
            decorator = ContractDecorator()
            decorator._origin_node = args[0]

            if len(args) > 1:
                values = self.validate_values(*args[:2])

                hash_arg = values[0] if len(values) > 0 else None
                if isinstance(hash_arg, UInt160):
                    decorator.contract_hash = hash_arg
            return decorator

        return self

    def validate_values(self, *params: Any) -> List[Any]:
        values = []
        if len(params) != 2:
            return values

        origin, visitor = params
        values.append(self.contract_hash)
        from boa3.analyser.astanalyser import IAstAnalyser
        if not isinstance(visitor, IAstAnalyser):
            return values

        from boa3.exception import CompilerError
        if not isinstance(origin, ast.Call) or len(origin.args) < 1:
            visitor._log_error(
                CompilerError.UnfilledArgument(origin.lineno, origin.col_offset, list(self.args.keys())[0])
            )
            return values
        argument_hash = visitor.visit(origin.args[0])

        try:
            if isinstance(argument_hash, str):
                from boa3.neo import from_hex_str
                argument_hash = from_hex_str(argument_hash)

            if isinstance(argument_hash, bytes):
                values[0] = UInt160(argument_hash)
        except BaseException:
            visitor._log_error(CompilerError.InvalidUsage(origin.lineno,
                                                          origin.col_offset,
                                                          "Only literal values are accepted for 'script_hash' argument"
                                                          ))

        return values

    @property
    def is_function_decorator(self) -> bool:
        return False

    @property
    def is_class_decorator(self) -> bool:
        return True



from typing import Any, Iterable, List, Set

from boa3.model.type.itype import IType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItem import StackItemType


class UnionType(IType):
    """
    An class used to represent Python Union annotation type
    """

    def __init__(self, union_types: Set[IType] = None):
        identifier: str = 'Union'
        super().__init__(identifier)
        self._union_types: Set[IType] = union_types

        # variables to not reevaluate everytime we need to access
        self._abi_type: AbiType = None
        self._stack_item: StackItemType = None

    @property
    def identifier(self) -> str:
        return '{0}[{1}]'.format(self._identifier,
                                 ', '.join([t.identifier for t in self._union_types]))

    @property
    def union_types(self) -> List[IType]:
        return list(self._union_types)

    @property
    def abi_type(self) -> AbiType:
        if self._abi_type is None:
            self._abi_type = AbiType.union([union.abi_type for union in self._union_types])
        return self._abi_type

    @property
    def stack_item(self) -> StackItemType:
        if self._stack_item is None:
            self._stack_item = StackItemType.union([union.stack_item for union in self._union_types])
        return self._stack_item

    def _is_type_of(self, value: Any) -> bool:
        if not isinstance(self._union_types, Iterable):
            return False
        if isinstance(value, UnionType):
            return all(self._is_type_of(x) for x in value._union_types)
        return any(t.is_type_of(value) for t in self._union_types)

    @classmethod
    def build(cls, value: Any) -> IType:
        if isinstance(value, Iterable):
            types = set(value)
        else:
            types = {value}

        from boa3.model.type.type import Type
        if any(t is Type.any for t in types):
            return Type.any

        if all(isinstance(t, IType) for t in types):
            for x in types.copy():
                if isinstance(x, UnionType):
                    types.remove(x)
                    types = types.union(x._union_types)
            if len(types) == 1:
                return list(types)[0]
            return cls(types)

    def except_type(self, other_type: IType) -> IType:
        """
        Gets a type that is type of `self` but is not type of `other_type`.
        If `other_type` is an implementation of `self`, returns `self`

        :param other_type: type that won't be allowed in the resulting type
        :type other_type: IType
        :return: an union of this type and the exclusion of other_type
        :rtype: IType
        """
        if not self.is_type_of(other_type):
            return self

        new_union = []
        for x in self._union_types:
            if not other_type.is_type_of(x):
                new_union.append(x)

        if len(new_union) < 2:
            from boa3.model.type.type import Type
            return Type.none if len(new_union) == 0 else new_union[0]
        else:
            return UnionType(set(new_union))

    def intersect_type(self, other_type: IType) -> IType:
        if self.is_type_of(other_type):
            return other_type
        if other_type.is_type_of(self):
            return self

        if isinstance(other_type, type(self)):
            other_types = [x for x in other_type.union_types if self.is_type_of(x)]
            self_types = [x for x in self.union_types if other_type.is_type_of(x)]

            common_types = other_types + self_types
            if len(common_types) == 1:
                return common_types[0]
            elif len(common_types) > 1:
                return self.build(common_types)

        return super().intersect_type(other_type)

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other) -> bool:
        return isinstance(other, UnionType) and self._union_types == other._union_types



from typing import Optional

from boa3.model.builtin.method.maxmethod import MaxMethod
from boa3.model.type.itype import IType


class MaxIntMethod(MaxMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        super().__init__(arg_value if Type.int.is_type_of(arg_value) else Type.int)



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ExceptionMethod(IBuiltinMethod):

    def __init__(self, argument_type: IType = None):
        from boa3.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.str

        identifier = '-Exception'
        args: Dict[str, Variable] = {'message': Variable(argument_type)}
        default_message = "'{0}'".format(self.default_message) if argument_type is Type.str else "{0}"
        default = ast.parse(default_message.format(argument_type.default_value)
                            ).body[0].value
        super().__init__(identifier, args, [default], return_type=Type.exception)

    @property
    def _arg_message(self) -> Variable:
        return self.args['message']

    @property
    def default_message(self) -> str:
        return 'usererror'

    @property
    def identifier(self) -> str:
        return self._identifier

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        return isinstance(param_type, PrimitiveType)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_message.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ExceptionMethod(value)
        return super().build(value)



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.collection.mapping.mutable.dicttype import DictType
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class JoinMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None, iterable_type: Union[SequenceType, DictType] = None):
        from boa3.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            self_type = ByteStringType.build()

        if not isinstance(iterable_type, (SequenceType, DictType)):
            iterable_type = Type.sequence.build_collection([self_type])

        identifier = 'join'
        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'iterable': Variable(iterable_type),
        }

        super().__init__(identifier, args, return_type=self_type)

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if Type.dict.is_type_of(self._arg_iterable.type):
            return '-{0}_{1}'.format(self._identifier, Type.dict.identifier)

        return self._identifier  # JoinMethod default value for self

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _arg_iterable(self) -> Variable:
        return self.args['iterable']

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')

        # Receives: string, iterable
        verify_empty_string = [         # verify if string is empty
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),         # iterable_size = len(iterable)
            (Opcode.PUSH0, b''),        # index = 0
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            jmp_place_holder            # if iterable_size is less than or equals to 0, return empty string
        ]

        initialize_string = [           # initializes the return value
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),     # joined = iterable[0]
            (Opcode.SWAP, b''),
            (Opcode.INC, b''),          # index++
            (Opcode.SWAP, b''),
        ]

        from boa3.model.type.type import Type
        if Type.dict.is_type_of(self._arg_iterable.type):
            get_dict_keys = [           # get dictionary keys as an array
                (Opcode.REVERSE3, b''),
                (Opcode.KEYS, b''),
                (Opcode.REVERSE3, b''),
            ]
            initialize_string = get_dict_keys + initialize_string

        verify_index = [                # verify if all items on the iterable were visited
            (Opcode.OVER, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            jmp_place_holder            # jump to remove_extra_values if every item was already visited
        ]

        concat_strings = [              # concatenate the items inside de interable
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.CAT, b''),          # joined = joined + string
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.CAT, b''),          # joined = joined + iterable[index]
            (Opcode.SWAP, b''),
            (Opcode.INC, b''),          # index++
            (Opcode.SWAP, b''),
            # jump back to verify_index
        ]

        jmp_back_to_verify = Opcode.get_jump_and_data(Opcode.JMP, -get_bytes_count(verify_index +
                                                                                   concat_strings))
        concat_strings.append(jmp_back_to_verify)

        jmp_concatenation = Opcode.get_jump_and_data(Opcode.JMPLE, get_bytes_count(initialize_string +
                                                                                   verify_index +
                                                                                   concat_strings), True)
        verify_empty_string[-1] = jmp_concatenation

        add_empty_string = [            # add a empty string at the top of the stack
            (Opcode.PUSHDATA1, b'\x00')
        ]

        jmp_concatenation = Opcode.get_jump_and_data(Opcode.JMPGE, get_bytes_count(concat_strings +
                                                                                   add_empty_string), True)
        verify_index[-1] = jmp_concatenation

        remove_extra_values = [         # remove all values from stack except the joined string
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
        ]

        return (
            verify_empty_string +
            initialize_string +
            verify_index +
            concat_strings +
            add_empty_string +
            remove_extra_values
        )

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, Iterable):
            value = [value]
        if isinstance(value, list) and len(value) <= 2:
            self_type = self._arg_self.type
            if len(value) < 2:
                value.append(None)
            if isinstance(value[0], type(self_type)) and isinstance(value[1], SequenceType):
                return self
            else:
                return JoinMethod(value[0], value[1])

        return super().build(value)



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3 import constants
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.uint256type import UInt256Type
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class UInt256Method(IBuiltinMethod):

    def __init__(self, return_type: UInt256Type, argument_type: IType = None):
        from boa3.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.none

        identifier = 'UInt256'
        args: Dict[str, Variable] = {'object': Variable(argument_type)}

        args_default = ast.parse("{0}".format(Type.int.default_value)
                                 ).body[0].value

        super().__init__(identifier, args, [args_default], return_type=return_type)

    @property
    def _arg_object(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_object.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_object.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.model.type.type import Type

        return (Type.bytes.is_type_of(param_type)
                or Type.int.is_type_of(param_type))

    def evaluate_literal(self, *args: Any) -> Any:
        from boa3.neo3.core.types import UInt256

        if len(args) == 0:
            return UInt256.zero().to_array()

        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, int):
                from boa3.neo.vm.type.Integer import Integer
                arg = Integer(arg).to_byte_array(min_length=UInt256._BYTE_LEN)
            if isinstance(arg, bytes):
                value = UInt256(arg).to_array()
                return value

        return super().evaluate_literal(*args)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        from boa3.model.type.type import Type

        from boa3.neo.vm.type.StackItem import StackItemType
        return [
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, Type.int.stack_item),  # if istype(arg, int):
            (Opcode.JMPIFNOT, Integer(56).to_byte_array(signed=True)),
            (Opcode.DUP, b''),                       # assert num >= 0
            (Opcode.PUSH0, b''),
            (Opcode.GE, b''),
            (Opcode.ASSERT, b''),

            (Opcode.DUP, b''),                       # if len(num) < 32
            (Opcode.SIZE, b''),                        # increase number's length to 32
            (Opcode.PUSHINT8, Integer(constants.SIZE_OF_INT256).to_byte_array(signed=True)),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),

            (Opcode.JMPGE, Integer(44).to_byte_array(signed=True)),
            (Opcode.PUSHDATA1, (Integer(constants.SIZE_OF_INT256).to_byte_array(signed=True)
                                + bytes(constants.SIZE_OF_INT256))),
            (Opcode.REVERSE3, b''),
            (Opcode.SUB, b''),
            (Opcode.LEFT, b''),
            (Opcode.CAT, b''),

            (Opcode.JMP, Integer(4).to_byte_array()),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),

            (Opcode.CONVERT, StackItemType.ByteString),  # convert to uint256
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSHINT8, Integer(constants.SIZE_OF_INT256).to_byte_array(signed=True)),
            (Opcode.NUMEQUAL, b''),
            (Opcode.ASSERT, b''),
        ]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_object.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return UInt256Method(self.return_type, value)
        return super().build(value)



import ast
from typing import Dict, List, Tuple

from boa3.model import set_internal_call
from boa3.model.builtin.classmethod.countmethod import CountMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class CountStrMethod(CountMethod):
    def __init__(self):
        from boa3.model.type.type import Type

        args: Dict[str, Variable] = {
            'self': Variable(Type.str),
            'value': Variable(Type.str),
            'start': Variable(Type.int),
            'end': Variable(Type.union.build([Type.int, Type.none])),
        }

        start_default = set_internal_call(ast.parse("{0}".format(Type.int.default_value)
                                                    ).body[0].value)
        end_default = set_internal_call(ast.parse("{0}".format(Type.none.default_value)
                                                  ).body[0].value)

        super().__init__(args, [start_default, end_default])

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.model.type.type import Type

        jmp_place_holder = (Opcode.JMP, b'\x01')

        # region verify end

        verify_end_null = [     # verifies if end is null
            (Opcode.REVERSE4, b''),
            (Opcode.DUP, b''),
            (Opcode.ISNULL, b''),
            jmp_place_holder
        ]

        end_null = [            # if end is null, end = len(self)
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder    # skip the other end verifications
        ]

        num_jmp_code = get_bytes_count(end_null)
        jmp_str_end_null_statement = Opcode.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
        verify_end_null[-1] = jmp_str_end_null_statement

        verify_end_neg = [      # verifies if end is < 0
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder
        ]

        end_neg = [             # if end is < 0, then get the positive equivalent number
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder
        ]

        end_still_neg = [       # if end is still < 0 after adding it to len(self), end = 0
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),
        ]

        num_jmp_code = get_bytes_count(end_still_neg)
        jmp_end_still_neg_statement = Opcode.get_jump_and_data(Opcode.JMPGE, num_jmp_code, True)
        end_neg[-1] = jmp_end_still_neg_statement

        skip_end_gt_size = [
            jmp_place_holder    # skip the other end verification
        ]

        num_jmp_code = get_bytes_count(end_still_neg + end_neg + skip_end_gt_size)
        jmp_end_neg_statement = Opcode.get_jump_and_data(Opcode.JMPGE, num_jmp_code, True)
        verify_end_neg[-1] = jmp_end_neg_statement

        verify_end_gt_size = [  # verifies if end >= len(self)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder,
        ]

        end_gt_size = [         # if end >= len(self), end = len(self)
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
        ]

        num_jmp_code = get_bytes_count(end_gt_size)
        jmp_end_gt_size_statement = Opcode.get_jump_and_data(Opcode.JMPLE, num_jmp_code, True)
        verify_end_gt_size[-1] = jmp_end_gt_size_statement

        num_jmp_code = get_bytes_count(end_gt_size + verify_end_gt_size)
        jmp_whole_end_gt_size_statement = Opcode.get_jump_and_data(Opcode.JMP, num_jmp_code, True)
        skip_end_gt_size[-1] = jmp_whole_end_gt_size_statement

        num_jmp_code = get_bytes_count(
            end_gt_size + verify_end_gt_size + skip_end_gt_size + end_still_neg + end_neg + verify_end_neg
        )
        jmp_whole_end_gt_size_statement = Opcode.get_jump_and_data(Opcode.JMP, num_jmp_code, True)
        end_null[-1] = jmp_whole_end_gt_size_statement

        verify_end = (
            verify_end_null +
            end_null +
            verify_end_neg +
            end_neg +
            end_still_neg +
            skip_end_gt_size +
            verify_end_gt_size +
            end_gt_size
        )

        # endregion

        # region verify start

        verify_start_neg = [    # verifies if start is < 0
            (Opcode.OVER, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder
        ]

        start_neg = [           # if start is < 0, then get the positive equivalent index
            (Opcode.SWAP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder
        ]

        start_still_neg = [     # if start is still < 0 after adding it to len(self), start = 0
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),
        ]

        num_jmp_code = get_bytes_count(start_still_neg)
        jmp_str_start_still_neg_statement = Opcode.get_jump_and_data(Opcode.JMPGE, num_jmp_code, True)
        start_neg[-1] = jmp_str_start_still_neg_statement

        correct_start_position = [  # put start on the correct position
            (Opcode.SWAP, b'')
        ]

        num_jmp_code = get_bytes_count(correct_start_position + start_still_neg + start_neg)
        jmp_str_start_neg_statement = Opcode.get_jump_and_data(Opcode.JMPGE, num_jmp_code, True)
        verify_start_neg[-1] = jmp_str_start_neg_statement

        verify_start = (
            verify_start_neg +
            start_neg +
            start_still_neg +
            correct_start_position
        )

        # endregion

        # region Count logic

        initialize = [          # initialize variables
            (Opcode.PUSH0, b''),  # count = 0
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),  # substr_size = len(substr)
            (Opcode.SWAP, b''),
        ]

        verify_while = [        # verifies if substr_size + index >= end
            (Opcode.OVER, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.ADD, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            jmp_place_holder
        ]

        count_substring = [     # verifies if self[index: index+substr_size] == substr
            (Opcode.PUSH5, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, Type.str.stack_item),
            (Opcode.PUSH5, b''),
            (Opcode.PICK, b''),
            (Opcode.NUMEQUAL, b''),
            jmp_place_holder
        ]

        count_plusplus = [      # if self[index: index+substr_size] == substr
            (Opcode.INC, b''),  # count++
        ]

        num_jmp_code = get_bytes_count(count_plusplus)
        jmp_count_plusplus_statement = Opcode.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
        count_substring[-1] = jmp_count_plusplus_statement

        go_back_to_while = [    # go back to while verification
            (Opcode.REVERSE4, b''),
            (Opcode.INC, b''),  # index ++
            (Opcode.REVERSE4, b''),
            # jump back to while
        ]

        num_jmp_code = -get_bytes_count(go_back_to_while + count_plusplus + count_substring + verify_while)
        jmp_back_to_while_statement = Opcode.get_jump_and_data(Opcode.JMP, num_jmp_code)
        go_back_to_while.append(jmp_back_to_while_statement)

        num_jmp_code = get_bytes_count(go_back_to_while + count_plusplus + count_substring)
        jmp_to_clean_stack_statement = Opcode.get_jump_and_data(Opcode.JMPGT, num_jmp_code, True)
        verify_while[-1] = jmp_to_clean_stack_statement

        clean_stack = [         # remove auxiliary values
            (Opcode.REVERSE4, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        count_logic = (
            initialize +
            verify_while +
            count_substring +
            count_plusplus +
            go_back_to_while +
            clean_stack
        )

        # endregion

        # count string logic is verify_end + verify_start  + count_logic
        return (
            verify_end +
            verify_start +
            count_logic
        )



from abc import ABC, abstractmethod


class ISymbol(ABC):
    defined_by_entry: bool = False

    @property
    @abstractmethod
    def shadowing_name(self) -> str:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        pass



__all__ = ['BalanceOfMethod',
           'DecimalsMethod',
           'SymbolMethod',
           'TotalSupplyMethod',
           'TransferMethod',
           ]

from boa3.model.builtin.native.nep17_methods.balanceofmethod import BalanceOfMethod
from boa3.model.builtin.native.nep17_methods.decimalsmethod import DecimalsMethod
from boa3.model.builtin.native.nep17_methods.symbolmethod import SymbolMethod
from boa3.model.builtin.native.nep17_methods.totalsupplymethod import TotalSupplyMethod
from boa3.model.builtin.native.nep17_methods.transfermethod import TransferMethod



from enum import IntEnum
from typing import Optional, Tuple

from boa3 import constants
from boa3.neo.vm.type.Integer import Integer


def get_native_contract_data(token_script: bytes) -> Tuple[Optional[bytes], Optional[int]]:
    prefix: Optional[NativeContractPrefix] = None
    contract_id: NativeContractId = NativeContractId.NONE

    if token_script is constants.NEO_SCRIPT:
        prefix = NativeContractPrefix.NEO
        contract_id = NativeContractId.NEO
    elif token_script is constants.GAS_SCRIPT:
        prefix = NativeContractPrefix.GAS
        contract_id = NativeContractId.GAS
    elif token_script is constants.MANAGEMENT_SCRIPT:
        prefix = NativeContractPrefix.ContractManagement
        contract_id = NativeContractId.ContractManagement

    return (prefix if prefix is None else Integer(prefix).to_byte_array(min_length=1),
            contract_id.value)


class NativeContractId(IntEnum):
    ContractManagement = -1
    NEO = -5
    GAS = -6
    Oracle = -9

    NONE = 0


class NativeContractPrefix(IntEnum):
    ContractManagement = 8
    NEO = 20
    GAS = 20

    NONE = 0



from typing import List, Optional

from boa3.model.operation.binary.additional import *
from boa3.model.operation.binary.arithmetic import *
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.binary.logical import *
from boa3.model.operation.binary.relational import *
from boa3.model.operation.operation import IOperation
from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.noneidentity import NoneIdentity
from boa3.model.operation.unary.nonenotidentity import NoneNotIdentity
from boa3.model.type.itype import IType


class BinaryOp:
    # Arithmetic operations
    Add = Addition()
    Sub = Subtraction()
    Mul = Multiplication()
    Div = Division()
    IntDiv = FloorDivision()
    ListAdd = ListAddition()
    Mod = Modulo()
    Pow = Power()
    Concat = Concat()
    StrBytesMul = StrBytesMultiplication()

    # Relational operations
    NumEq = NumericEquality()
    NumNotEq = NumericInequality()
    Lt = LessThan()
    LtE = LessThanOrEqual()
    Gt = GreaterThan()
    GtE = GreaterThanOrEqual()
    IsNone = NoneIdentity()
    IsNotNone = NoneNotIdentity()
    Is = Identity()
    IsNot = NotIdentity()
    Eq = ObjectEquality()
    NotEq = ObjectInequality()

    # Logical operations
    And = BooleanAnd()
    Or = BooleanOr()
    BitAnd = LogicAnd()
    BitOr = LogicOr()
    Xor = LogicXor()
    LShift = LeftShift()
    RShift = RightShift()

    # Other operations
    In = CollectionMembership()
    NotIn = CollectionNotMembership()

    @classmethod
    def validate_type(cls, operator: Operator, left: IType, right: IType) -> Optional[BinaryOperation]:
        """
        Gets a binary operation given the operator and the operands types.

        :param operator: binary operator
        :param left: type of the left operand
        :param right: type of the right operand
        :return: The operation if exists. None otherwise;
        :rtype: BinaryOperation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, IOperation) and op.is_valid(operator, left, right):
                if isinstance(op, BinaryOperation):
                    return op.build(left, right)
                else:
                    from boa3.model.type.type import Type
                    operand = right if left is Type.none else left
                    return op.build(operand)

    @classmethod
    def get_operation_by_operator(cls, operator: Operator, left_operand: IType,
                                  right_operand: Optional[IType] = None) -> Optional[BinaryOperation]:
        """
        Gets a binary operation given the operator.

        :param operator: binary operator
        :param left_operand: left operand of the operator
        :param right_operand: right operand of the operator
        :return: The operation if exists. If exists more than one operation with the same operator, returns the one with
        the same left operand. If none has the same left operand, returns the first found. None otherwise;
        :rtype: BinaryOperation or None
        """
        valid_operations: List[BinaryOperation] = []
        for id, op in vars(cls).items():
            if isinstance(op, BinaryOperation) and op.operator is operator:
                left, right = op.get_valid_operand_for_validation(left_operand, right_operand)
                if left is not None:
                    return op.build(left_operand if right_operand is None else left, right)
                else:
                    valid_operations.append(op)

        return valid_operations[0] if len(valid_operations) > 0 else None

    @classmethod
    def get_operation(cls, operation: BinaryOperation) -> Optional[BinaryOperation]:
        """
        Gets a binary operation given another operation.

        :param operation: binary operation
        :return: The operation if exists. None otherwise;
        :rtype: BinaryOperation or None
        """
        for id, op in vars(cls).items():
            if type(operation) == type(op):
                return op



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class StorageMapDeleteMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.builtin.interop.storage.storagemap.storagemaptype import _StorageMap
        from boa3.model.type.type import Type
        from boa3.model.type.primitive.bytestringtype import ByteStringType

        identifier = 'delete'
        byte_string_type = ByteStringType.build()

        args: Dict[str, Variable] = {'self': Variable(_StorageMap),
                                     'key': Variable(byte_string_type)}

        super().__init__(identifier, args, return_type=Type.none)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.builtin.interop.interop import Interop
        return [
            (Opcode.SWAP, b''),
            (Opcode.OVER, b''),
            (Opcode.PUSH1, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.SWAP, b''),
            (Opcode.CAT, b''),
            (Opcode.SWAP, b''),
            (Opcode.PUSH0, b''),
            (Opcode.PICKITEM, b''),
        ] + Interop.StorageDelete.opcode



from typing import Any

from boa3.model.type.itype import IType
from boa3.model.type.primitive.ibytestringtype import IByteStringType


class ByteStringType(IByteStringType):
    """
    A class used to represent ByteString type interface
    """

    def __init__(self):
        identifier = 'ByteString'
        values_type = [self]
        super().__init__(identifier, values_type)

    @classmethod
    def _is_type_of(cls, value: Any) -> bool:
        from boa3.model.type.type import Type
        return (Type.str.is_type_of(value)
                or Type.bytes.is_type_of(value)
                or isinstance(value, ByteStringType))

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.model.builtin.builtin import Builtin

        instance_methods = [Builtin.ConvertToBytes,
                            Builtin.ConvertToBool,
                            Builtin.ConvertToInt,
                            Builtin.ConvertToStr,
                            ]

        for instance_method in instance_methods:
            self._instance_methods[instance_method.raw_identifier] = instance_method.build(self)

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _ByteString


_ByteString = ByteStringType()



from boa3.neo3.contracts.contracttypes import CallFlags, TriggerType
from boa3.neo3.contracts.findoptions import FindOptions

__all__ = ['CallFlags',
           'FindOptions',
           'TriggerType']



from __future__ import annotations

from enum import Enum
from typing import Any, List

from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String


class StackItemType(bytes, Enum):
    Any = b'\x00'
    Pointer = b'\x10'
    Boolean = b'\x20'
    Integer = b'\x21'
    ByteString = b'\x28'
    Buffer = b'\x30'
    Array = b'\x40'
    Struct = b'\x41'
    Map = b'\x48'
    InteropInterface = b'\x60'

    @staticmethod
    def get_stack_item_type(stack_item_type: str) -> StackItemType:
        try:
            return StackItemType[stack_item_type]
        except BaseException:
            return StackItemType.Any

    @staticmethod
    def union(stack_item_types: List[StackItemType]) -> StackItemType:
        if len(stack_item_types) == 0:
            return StackItemType.Any

        if len(stack_item_types) == 1:
            return stack_item_types[0]

        stacks = stack_item_types[:1]
        for union in stack_item_types[1:]:
            if union not in stacks:
                stacks.append(union)

        if len(stacks) == 1:
            return stacks[0]
        else:
            return StackItemType.Any


def serialize(value: Any) -> bytes:
    if value is None:
        return StackItemType.Any

    if isinstance(value, bool):
        stack_type = StackItemType.Boolean
        span = Integer(value).to_byte_array(signed=False, min_length=1)
        return stack_type + span

    if isinstance(value, (int, str, bytes)):
        if isinstance(value, int):
            stack_type = StackItemType.Integer
            span = Integer(value).to_byte_array(signed=True, min_length=1)
        elif isinstance(value, str):
            stack_type = StackItemType.ByteString
            span = String(value).to_bytes()
        else:
            stack_type = StackItemType.Buffer
            span = value

        return (stack_type
                + Integer(len(span)).to_byte_array(signed=False, min_length=1)
                + span)

    if isinstance(value, (list, tuple, set, range)):
        serialized = StackItemType.Array + Integer(len(value)).to_byte_array(signed=False)
        for x in value:
            serialized += serialize(x)
        return serialized

    if isinstance(value, dict):
        serialized = StackItemType.Map + Integer(len(value)).to_byte_array(signed=False)
        for key, x in value.items():
            serialized += serialize(key)
            serialized += serialize(x)
        return serialized

    return b''



from typing import Any, Dict, List, Optional, Sized, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ScriptHashMethod(IBuiltinMethod):

    def __init__(self, data_type: IType = None):
        if (not Type.int.is_type_of(data_type)
                and not Type.str.is_type_of(data_type)
                and not Type.bytes.is_type_of(data_type)):
            data_type = Type.any

        identifier = 'to_script_hash'
        args: Dict[str, Variable] = {'self': Variable(data_type)}
        super().__init__(identifier, args, return_type=Type.bytes)

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        self_type = self.args['self'].type
        if self_type is Type.any:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self_type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, PrimitiveType)

    @property
    def is_supported(self) -> bool:
        return self.args['self'].type is not Type.any

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.constants import SIZE_OF_INT160
        from boa3.model.builtin.interop.stdlib.base58decodemethod import Base58DecodeMethod
        from boa3.model.type.type import Type
        from boa3.neo.vm.type.Integer import Integer

        opcodes = [
            (Opcode.DUP, b''),      # convert value to string
            (Opcode.SIZE, b''),
            (Opcode.JMPIFNOT, Integer(36).to_byte_array(signed=True, min_length=1)),
            (Opcode.DUP, b''),      # convert value to string
            (Opcode.ISTYPE, Type.str.stack_item),
            (Opcode.JMPIF, Integer(4).to_byte_array(min_length=1)),
            (Opcode.CONVERT, Type.str.stack_item),
            (Opcode.PUSH1, b''),
            (Opcode.PACK, b'')
        ]
        opcodes.extend(Base58DecodeMethod().opcode)
        script_len = Integer(SIZE_OF_INT160).to_byte_array(min_length=1)
        opcodes.extend([
            (Opcode.DUP, b''),      # if len(result) > SIZE_OF_INT160, truncates the result
            (Opcode.SIZE, b''),
            (Opcode.PUSHDATA1, Integer(len(script_len)).to_byte_array(min_length=1) + script_len),
            (Opcode.CONVERT, Type.int.stack_item),
            (Opcode.JMPGT, Integer(8).to_byte_array(min_length=1, signed=True)),
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),     # first byte identifies address version
            (Opcode.DEC, b''),
            (Opcode.RIGHT, b''),
            (Opcode.JMP, Integer(9).to_byte_array(min_length=1, signed=True)),
            (Opcode.PUSH1, b''),
            (Opcode.PUSHDATA1, Integer(len(script_len)).to_byte_array(min_length=1) + script_len),
            (Opcode.CONVERT, Type.int.stack_item),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, Type.bytes.stack_item)
        ])
        return opcodes

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if 'self' in self.args and self.args['self'].type is not Type.any:
            return self

        from boa3.model.type.collection.sequence.ecpointtype import ECPointType
        from boa3.model.type.primitive.inttype import IntType
        from boa3.model.type.primitive.strtype import StrType
        from boa3.model.type.primitive.bytestype import BytesType

        if isinstance(value, Sized) and len(value) == 1:
            value = value[0]

        if isinstance(value, ECPointType):
            from boa3.model.builtin.method.ecpointtoscripthashmethod import ECPointToScriptHashMethod
            return ECPointToScriptHashMethod()
        elif isinstance(value, (IntType, StrType, BytesType)):
            return ScriptHashMethod(value)
        elif isinstance(value, IType):
            return ScriptHashMethod(Type.bytes)
        return super().build(value)



from typing import Union

from boa3.builtin import CreateNewEvent, Event
from boa3.builtin.type import ByteString, ECPoint, UInt160

Nep5TransferEvent: Event = CreateNewEvent(
    [
        ('from', bytes),
        ('to', bytes),
        ('amount', int)
    ],
    'transfer'
)
"""
The NEP-5 transfer event that will be triggered whenever a token is transferred, minted or burned. It needs the 
addresses of the sender, receiver and the amount transferred.

:meta hide-value:
"""

Nep11TransferEvent: Event = CreateNewEvent(
    [
        ('from', Union[UInt160, None]),
        ('to', Union[UInt160, None]),
        ('amount', int),
        ('tokenId', ByteString)
    ],
    'Transfer'
)
"""
The NEP-11 Transfer event that will be triggered whenever a token is transferred, minted or burned. It needs the 
addresses of the sender, receiver, amount transferred and the id of the token.

:meta hide-value:
"""


Nep17TransferEvent: Event = CreateNewEvent(
    [
        ('from', Union[UInt160, None]),
        ('to', Union[UInt160, None]),
        ('amount', int)
    ],
    'Transfer'
)
"""
The NEP-17 Transfer event that will be triggered whenever a token is transferred, minted or burned. It needs the 
addresses of the sender, receiver and the amount transferred.

:meta hide-value:
"""


def abort():
    """
    Aborts the execution of a smart contract.
    """
    pass


class NeoAccountState:
    """
    Represents the account state of NEO token in the NEO system.

    :ivar balance: the current account balance, which equals to the votes cast
    :vartype balance: int
    :ivar height: the height of the block where the balance changed last time
    :vartype height: int
    :ivar vote_to: the voting target of the account
    :vartype vote_to: ECPoint
    """

    def __init__(self):
        from boa3 import constants
        self.balance: int = 0
        self.height: int = 0
        self.vote_to: ECPoint = ECPoint(bytes(constants.SIZE_OF_ECPOINT))



import ast
from typing import Any, Dict, List, Sized

from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.variable import Variable


class DisplayNameDecorator(IBuiltinDecorator):
    def __init__(self):
        from boa3.model.type.type import Type

        identifier = 'display_name'
        args: Dict[str, Variable] = {'name': Variable(Type.str)}
        super().__init__(identifier, args)
        self.external_name = None

    def build(self, *args) -> IBuiltinDecorator:
        if isinstance(args, Sized) and len(args) > 0 and isinstance(args[0], ast.AST):
            decorator = DisplayNameDecorator()
            decorator._origin_node = args[0]

            if len(args) > 1:
                values = self.validate_values(*args[:2])

                hash_arg = values[0] if len(values) > 0 else None
                if isinstance(hash_arg, str):
                    decorator.external_name = hash_arg
            return decorator

        return self

    def validate_values(self, *params: Any) -> List[Any]:
        values = []
        if len(params) != 2:
            return values

        origin, visitor = params
        values.append(self.external_name)
        from boa3.analyser.astanalyser import IAstAnalyser
        if not isinstance(visitor, IAstAnalyser):
            return values

        from boa3.exception import CompilerError
        if not isinstance(origin, ast.Call):
            visitor._log_error(
                CompilerError.UnfilledArgument(origin.lineno, origin.col_offset, list(self.args.keys())[0])
            )
            return values

        args_names = list(self.args.keys())
        if len(origin.args) > len(args_names):
            visitor._log_error(
                CompilerError.UnexpectedArgument(origin.lineno, origin.col_offset)
            )
            return values

        # read the called arguments
        args_len = min(len(origin.args), len(args_names))
        values.clear()

        for x in range(args_len):
            values.append(visitor.visit(origin.args[x]))

        if len(values) < 1:
            values.append(self.external_name)

        # read the called keyword arguments
        for kwarg in origin.keywords:
            if kwarg.arg in args_names:
                x = args_names.index(kwarg.arg)
                value = visitor.visit(kwarg.value)
                values[x] = value
            else:
                visitor._log_error(
                    CompilerError.UnexpectedArgument(kwarg.lineno, kwarg.col_offset)
                )

        if not isinstance(values[0], str):
            visitor._log_error(
                CompilerError.UnfilledArgument(origin.lineno, origin.col_offset, list(self.args.keys())[0])
            )
            return values

        if visitor.get_symbol(values[0]) is not None:
            visitor._log_error(CompilerError.InvalidUsage(origin.lineno,
                                                          origin.col_offset,
                                                          "Only literal values are accepted for 'name' argument"
                                                          ))
            values[0] = self.external_name

        return values

    @property
    def is_function_decorator(self) -> bool:
        return False

    @property
    def is_class_decorator(self) -> bool:
        return True



from boa3.builtin.type import UInt160, UInt256


class Transaction:
    """
    Represents a transaction.

    :ivar hash: a unique identifier based on the unsigned data portion of the object
    :vartype hash: UInt256
    :ivar version: the data structure version of the transaction
    :vartype version: int
    :ivar nonce: a random number used once in the cryptography
    :vartype nonce: int
    :ivar sender: the sender is the first signer of the transaction, they will pay the fees of the transaction
    :vartype sender: UInt160
    :ivar system_fee: the fee paid for executing the `script`
    :vartype system_fee: int
    :ivar network_fee: the fee paid for the validation and inclusion of the transaction in a block by the consensus node
    :vartype network_fee: int
    :ivar valid_until_block: indicates that the transaction is only valid before this block height
    :vartype valid_until_block: int
    :ivar script: the array of instructions to be executed on the transaction chain by the virtual machine
    :vartype script: bytes
    """

    def __init__(self):
        self.hash: UInt256 = UInt256()
        self.version: int = 0
        self.nonce: int = 0
        self.sender: UInt160 = UInt160()
        self.system_fee: int = 0
        self.network_fee: int = 0
        self.valid_until_block: int = 0
        self.script: bytes = b''



from typing import Dict, Optional, Tuple

from boa3.model.variable import Variable


class FunctionArguments:

    def __init__(self):
        self._args: Dict[str, Variable] = {}
        self._vararg: Optional[Tuple[str, Variable]] = None
        self._kwargs: Optional[Dict[str, Variable]] = None

    @property
    def args(self) -> Dict[str, Variable]:
        return self._args.copy()

    def add_arg(self, arg_id: str, arg: Variable) -> bool:
        if not isinstance(arg, Variable):
            return False
        self._args[arg_id] = arg
        return True

    @property
    def vararg(self) -> Optional[Tuple[str, Variable]]:
        return self._vararg

    def set_vararg(self, arg_id: str, arg: Variable) -> bool:
        if not isinstance(arg, Variable):
            return False
        self._vararg = (arg_id, arg)
        return True

    @property
    def kwargs(self) -> Dict[str, Variable]:
        return self._kwargs

    def add_kwarg(self, arg_id: str, arg: Variable) -> bool:
        if not isinstance(arg, Variable):
            return False
        if self._kwargs is None:
            self._kwargs: Dict[str, Variable] = {}
        self._kwargs[arg_id] = arg
        return True



from __future__ import annotations

from typing import Any, Dict, List

from boa3_test.tests.test_classes.contract.neoeventstruct import NeoEventStruct
from boa3_test.tests.test_classes.contract.neomethodstruct import NeoMethodStruct
from boa3_test.tests.test_classes.contract.neostruct import NeoStruct


class NeoAbiStruct(NeoStruct):
    _methods_field = 'methods'
    _events_field = 'events'

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> NeoAbiStruct:
        required_fields = [cls._methods_field,
                           cls._events_field
                           ]
        cls._validate_json(json, required_fields)

        struct = cls()
        struct.append(cls._get_methods(json[cls._methods_field]))
        struct.append(cls._get_events(json[cls._events_field]))

        return struct

    @classmethod
    def _get_methods(cls, json_list: List[Dict[str, Any]]) -> List[NeoMethodStruct]:
        return [NeoMethodStruct.from_json(method_json) for method_json in json_list]

    @classmethod
    def _get_events(cls, json_list: List[Dict[str, Any]]) -> List[NeoEventStruct]:
        return [NeoEventStruct.from_json(method_json) for method_json in json_list]



from typing import Dict

from boa3.model.builtin.interop.nativecontract import PolicyContractMethod
from boa3.model.variable import Variable


class GetStoragePriceMethod(PolicyContractMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'get_storage_price'
        native_identifier = 'getStoragePrice'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args, return_type=Type.int)



from __future__ import annotations

import ast
from enum import Enum
from typing import Dict, Optional, Type


class Operator(str, Enum):
    # Arithmetic operators
    Plus = '+'
    Minus = '-'
    Mult = '*'
    Div = '/'
    IntDiv = '//'
    Mod = '%'
    Pow = '**'

    # Relational operators
    Eq = '=='
    NotEq = '!='
    Lt = '<'
    LtE = '<='
    Gt = '>'
    GtE = '>='
    Is = 'is'
    IsNot = 'is not'

    # Logical operators
    And = 'and'
    Or = 'or'
    Not = 'not'
    BitAnd = '&'
    BitOr = '|'
    BitNot = '~'
    BitXor = '^'
    LeftShift = '<<'
    RightShift = '>>'

    # Other operators
    Subscript = '[]'
    In = 'in'
    NotIn = 'not in'

    @classmethod
    def get_operation(cls, node: ast.operator) -> Optional[Operator]:
        operators: Dict[Type[ast.operator], Operator] = {
            ast.Add: Operator.Plus,
            ast.Sub: Operator.Minus,
            ast.Mult: Operator.Mult,
            ast.Div: Operator.Div,
            ast.FloorDiv: Operator.IntDiv,
            ast.Mod: Operator.Mod,
            ast.Pow: Operator.Pow,
            ast.UAdd: Operator.Plus,
            ast.USub: Operator.Minus,
            ast.Eq: Operator.Eq,
            ast.NotEq: Operator.NotEq,
            ast.Lt: Operator.Lt,
            ast.LtE: Operator.LtE,
            ast.Gt: Operator.Gt,
            ast.GtE: Operator.GtE,
            ast.Is: Operator.Is,
            ast.IsNot: Operator.IsNot,
            ast.In: Operator.In,
            ast.NotIn: Operator.NotIn,
            ast.And: Operator.And,
            ast.Or: Operator.Or,
            ast.Not: Operator.Not,
            ast.BitAnd: Operator.BitAnd,
            ast.BitOr: Operator.BitOr,
            ast.BitXor: Operator.BitXor,
            ast.Invert: Operator.BitNot,
            ast.LShift: Operator.LeftShift,
            ast.RShift: Operator.RightShift
        }

        node_type = type(node)
        if node_type in operators:
            return operators[node_type]
        else:
            return None

    def requires_right_operand_for_validation(self) -> bool:
        return self in [Operator.In,
                        Operator.NotIn
                        ]

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class CheckSigMethod(InteropMethod):

    _RAW_BYTES = b''

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'check_sig'
        syscall = 'System.Crypto.CheckSig'
        args: Dict[str, Variable] = {
            'pubkeys': Variable(ECPointType.build()),
            'signatures': Variable(Type.bytes)
        }
        super().__init__(identifier, syscall, args, return_type=Type.bool)

    @classmethod
    def get_raw_bytes(cls) -> bytes:
        if len(cls._RAW_BYTES) == 0:
            from boa3.compiler.codegenerator import get_bytes
            check_sig = CheckSigMethod()
            cls._RAW_BYTES = get_bytes(check_sig.opcode)

        return cls._RAW_BYTES



from typing import Dict

from boa3.model.builtin.interop.nativecontract import StdLibMethod
from boa3.model.variable import Variable


class Base58CheckEncodeMethod(StdLibMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'base58_check_encode'
        native_identifier = 'base58CheckEncode'
        args: Dict[str, Variable] = {'key': Variable(Type.bytes)}
        super().__init__(identifier, native_identifier, args, return_type=Type.str)



__all__ = ['GetExecFeeFactorMethod',
           'GetFeePerByteMethod',
           'GetStoragePriceMethod',
           'IsBlockedMethod'
           ]

from boa3.model.builtin.interop.policy.getexecfeefactormethod import GetExecFeeFactorMethod
from boa3.model.builtin.interop.policy.getfeeperbytemethod import GetFeePerByteMethod
from boa3.model.builtin.interop.policy.getstoragepricemethod import GetStoragePriceMethod
from boa3.model.builtin.interop.policy.isblockedmethod import IsBlockedMethod



from __future__ import annotations

from typing import Any, Dict

from boa3_test.tests.test_classes.contract.neostruct import NeoStruct


class NeoEventStruct(NeoStruct):
    _name_field = 'name'
    _parameters_field = 'parameters'

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> NeoEventStruct:
        required_fields = [cls._name_field,
                           cls._parameters_field
                           ]
        cls._validate_json(json, required_fields)

        struct = cls()
        struct.append(json[cls._name_field])
        struct.append([cls._get_param_info(arg) for arg in json[cls._parameters_field]])

        return struct



from abc import ABC
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.type.itype import IType
from boa3.model.type.primitive.bytestringtype import ByteStringType
from boa3.model.type.primitive.bytestype import BytesType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ToIntMethod(IBuiltinMethod, ABC):
    def __init__(self, self_type: IType):
        identifier = 'to_int'
        if isinstance(self_type, IdentifiedSymbol):
            identifier = '-{0}_{1}'.format(self_type.identifier, identifier)

        args: Dict[str, Variable] = {'self': Variable(self_type)}
        from boa3.model.type.type import Type
        super().__init__(identifier, args, return_type=Type.int)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        return isinstance(params[0], IExpression) and isinstance(params[0].type, BytesType)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.type import Type
        return [
            (Opcode.CONVERT, Type.int.stack_item)
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None


class _ConvertToIntMethod(ToIntMethod):
    def __init__(self):
        super().__init__(None)

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, (BytesType, ByteStringType)):
            return BytesToIntMethod(value)
        # if it is not a valid type, show mismatched type with bytes
        return BytesToIntMethod()


ToInt = _ConvertToIntMethod()


class BytesToIntMethod(ToIntMethod):
    def __init__(self, self_type: IType = None):
        if not isinstance(self_type, (BytesType, ByteStringType)):
            from boa3.model.type.type import Type
            self_type = Type.bytes
        super().__init__(self_type)

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, BytesType):
            return BytesToIntMethod(value)
        return super().build(value)



from __future__ import annotations

import ast
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class IBuiltinMethod(IBuiltinCallable, Method, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None):
        super().__init__(identifier, args, vararg, kwargs, defaults, return_type)

    @property
    def is_supported(self) -> bool:
        """
        Verifies if the builtin method is supported by the compiler

        :return: True if it is supported. False otherwise.
        """
        return True

    def not_supported_str(self, callable_id: str) -> str:
        return '{0}({1})'.format(callable_id,
                                 ','.join([arg.type.identifier for arg in self.args.values()]))

    @property
    def is_cast(self) -> bool:
        """
        Returns whether this method is for casting types
        """
        return False

    @property
    def cast_types(self) -> Optional[Tuple[IType, IType]]:
        """
        If `is_cast` is True, must return the original type and the target type of the cast.
        Otherwise, must return None
        """
        return None

    @property
    def args_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments if opcode is not empty. Zero otherwise.
        """
        if len(self.opcode) > 0:
            if self.pack_arguments:
                return 1

            num_args = self._args_on_stack
            if num_args < 0:
                return 0
            elif num_args > len(self.args):
                return len(self.args)
            return num_args
        else:
            return 0

    def push_self_first(self) -> bool:
        """
        Verifies if the `self` value of the method needs to be pushed to the Neo execution stack before the
        other arguments.

        :return: a boolean value indicating if the `self` argument must be pushed before. Returns False if there isn't
                 a `self` argument in the function
        """
        return False

    def validate_self(self, self_type: IType) -> bool:
        """
        Verifies if the given value is valid to the function `self` argument

        :param self_type: type of the value
        :return: a boolean value that represents if the value is valid. Returns False if there isn't a `self` argument
                 in the function
        """
        if not self.has_self_argument:
            return False
        return self.args['self'].type.is_type_of(self_type)

    @property
    def has_self_argument(self) -> bool:
        """
        Verifies if the function has a `self` argument.

        :return: True if there is this argument. False otherwise.
        """
        return 'self' in self.args

    @property
    @abstractmethod
    def _args_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments.
        """
        return 0

    @property
    def stores_on_slot(self) -> bool:
        """
        Returns whether this method needs to update the value from a variable

        :return: whether the method needs to update a variable
        """
        return False

    @property
    def requires_reordering(self) -> bool:
        """
        Returns whether this method requires a reordering of user inputted parameters

        :return: whether the method requires the parameters to be ordered differently as inputted from the user
        """
        return False

    def reorder(self, arguments: list):
        """
        Reorder the arguments if the method requires it.

        If `requires_reordering` returns True, this method must be implemented

        :param arguments: list of arguments to be reordered
        """
        pass

    @property
    def generation_order(self) -> List[int]:
        """
        Gets the indexes order that need to be used during code generation.
        If the order for generation is the same as inputted in code, returns reversed(range(0,len_args))

        :return: Index order for code generation
        """
        indexes = list(reversed(range(0, len(self.args))))
        if self.push_self_first():
            indexes.remove(0)     # remove the first index from whatever position it is
            indexes.insert(0, 0)  # and move to the first position

        return indexes

    def validate_negative_arguments(self) -> List[int]:
        """
        Returns a list of the arguments that have to be positive values and need validation.

        :return: list with the arguments indexes that need to be fixed.
        :rtype: List[int]
        """
        return []

    @property
    def pack_arguments(self) -> bool:
        """
        Return whether this method requires its parameters to be packed into an array

        :return: whether this method requires an array
        """
        return False

    def evaluate_literal(self, *args: Any) -> Any:
        """
        Tries to evaluate the result during compile time. If it cannot be evaluated, returns Undefined.

        :return: arguments to try to run the method
        :rtype: Any
        """
        from boa3.analyser.model.optimizer import Undefined
        return Undefined

    @property
    def body(self) -> Optional[str]:
        """
        Gets the body of the method.

        :return: Return the code of the method body if there is no opcode. None otherwise.
        """
        return self._body if len(self.opcode) <= 0 else None

    @property
    @abstractmethod
    def _body(self) -> Optional[str]:
        """
        Gets the body of the method.

        :return: Return the code of the method body.
        """
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        """
        Creates a method instance with the given value as self

        :param value: value to build the type
        :return: The built method if the value is valid. The current object otherwise
        :rtype: IBuiltinMethod
        """
        return self
